import random
import string
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from numpy import ndarray
from torch import Tensor


def set_random_seed(seed: int = None) -> None:
    r"""Set random seed for random, numpy, and pytorch.

    Args:
      seed(int): The random seed, defaults to none.

    """
    if seed is not None:
        seed = int(seed)

    # Set random seed for random
    random.seed(seed)
    # Set random seed for all numpy operations
    np.random.seed(seed=seed)

    # Set random seed for PyTorch
    if isinstance(seed, int):
        torch.manual_seed(seed)


def list_to_tensor(
    item_list: List[Tensor], min_len: int = 0, max_len: int = 0
) -> Tuple[Tensor, Optional[Tensor]]:
    r"""Convert the list of torch.Tensor into a torch.Tensor.

    Args:
      item_list(List[Tensor]): The tensor for converting.
      min_len(int): Min length of sequence of data, defaults to 0.
      max_len(int): Max length of sequence of data, defaults to 0.

    Returns:
      Tuple[Tensor, Optional[Tensor]]: The converted tensor and the corresponding
        mask tensor.

    """

    item_mask_tensor = None

    # Convert single value tensor
    if all(item_list[i].dim() == 0 for i in range(len(item_list))):
        item_tensor = torch.stack(item_list, dim=0)
    # Convert 2 or more-D tensor with the same shape
    elif all(
        (item_list[i].size() == item_list[0].size()) and (len(item_list[i].size()) != 1)
        for i in range(len(item_list))
    ):
        item_tensor = torch.stack(item_list, dim=0)
    # Convert reshape to 1-D tensor and then convert
    else:
        item_tensor, item_mask_tensor = pad_batch(
            [item.view(-1) for item in item_list], min_len, max_len
        )

    return item_tensor, item_mask_tensor


def pad_batch(
    batch: List[Tensor],
    min_len: int = 0,
    max_len: int = 0,
    pad_value: int = 0,
    left_padded: bool = False,
) -> Tuple[Tensor, Tensor]:
    r"""Convert the batch into a padded tensor and mask tensor.

    Args:
      batch(List[Tensor]): The tensor for padding.
      min_len(int): Min length of sequence of padding, defaults to 0.
      max_len(int): Max length of sequence of padding, defaults to 0.
      pad_value(int): The value to use for padding, defaults to 0.
      left_padding(boolean): If True, pad on the left, otherwise on the right,
        defaults to False.

    Returns:
      Tuple[Tensor, Tensor]: The padded tensor and corresponding mask tensor.

    """
    batch_size = len(batch)
    max_seq_len = int(np.max([item.size()[0] for item in batch]))

    if max_len > 0 and max_len < max_seq_len:
        max_seq_len = max_len

    max_seq_len = max(max_seq_len, min_len)

    padded_batch = batch[0].new_full((batch_size, max_seq_len), pad_value)

    for i, item in enumerate(batch):
        length = min(item.size()[0], max_seq_len)
        if left_padded:
            padded_batch[i, -length:] = item[-length:]
        else:
            padded_batch[i, :length] = item[:length]

    mask_batch = torch.eq(padded_batch.clone().detach(), pad_value)

    return padded_batch, mask_batch


def prob_to_pred(probs: ndarray) -> ndarray:
    r"""Identify the class with the maximum probability.

    Args:
      probs(ndarray): predicted probabilities.

    Returns:
      ndarray: predicted labels.

    """

    return np.argmax(probs, axis=-1)


def pred_to_prob(preds: ndarray, n_classes: int) -> ndarray:
    r"""Converts predicted labels to probabilistic labels.

    Args:
      preds(ndarray): Predicted labels.
      n_classes(int): Total number of classes.

    Returns:
      ndarray: predicted probabilities.

    """

    preds = preds.reshape(-1)
    probs = np.zeros((preds.shape[0], n_classes))

    for idx, class_idx in enumerate(preds):
        probs[idx, class_idx] = 1.0

    return probs


def move_to_device(obj: Any, device: Optional[int] = -1) -> Any:
    r"""Given a structure (possibly) containing Tensors on the CPU, move all the Tensors
      to the specified GPU (or do nothing, if they should beon the CPU).

        device = -1 -> "cpu"
        device =  0 -> "cuda:0"

      Originally from:
        https://github.com/HazyResearch/metal/blob/mmtl_clean/metal/utils.py

    Args:
      obj(Any): The object to convert.
      device(int): The device id, defaults to -1.

    Returns:
      Any: The converted object.

    """

    if device < 0 or not torch.cuda.is_available():
        return obj
    elif isinstance(obj, torch.Tensor):
        return obj.cuda(device)  # type: ignore
    elif isinstance(obj, dict):
        return {key: move_to_device(value, device) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [move_to_device(item, device) for item in obj]
    elif isinstance(obj, tuple):
        return tuple([move_to_device(item, device) for item in obj])
    else:
        return obj


def array_to_numpy(
    array: Union[ndarray, List[Any], Tensor], flatten: bool = False
) -> ndarray:
    r"""Covert an array to a numpy array.

    Args:
      array(ndarray or list or Tensor): An array to convert.
      flatten(bool): Whether to flatten or not.

    Returns:
      ndarray: Converted array.

    """

    if isinstance(array, np.ndarray):
        pass
    elif isinstance(array, list):
        array = np.array(array)
    elif isinstance(array, torch.Tensor):
        array = array.cpu().numpy()
    else:
        raise ValueError(f"Unrecognized type {type(array)} to convert to ndarray")

    if flatten:
        array = array.reshape(-1)

    return array


def merge(
    x: Dict[str, Any], y: Dict[str, Any], specical_keys: Union[str, List[str]] = None
) -> Dict[str, Any]:
    r"""Merge two nested dictionaries. Overwrite values in x with values in y.

    Args:
      x(dict): The original dict.
      y(dict): The new dict.
      specical_keys(str or list of str): The specical keys to replace
        instead of merging, defaults to None.

    Returns:
      dict: The updated dic.

    """

    if x is None:
        return y
    if y is None:
        return x

    if isinstance(specical_keys, str):
        specical_keys = [specical_keys]

    merged = {**x, **y}

    xkeys = x.keys()

    for key in xkeys:
        if specical_keys is not None and key in specical_keys and key in y:
            merged[key] = y[key]
        elif isinstance(x[key], dict) and key in y:
            merged[key] = merge(x[key], y[key], specical_keys)

    return merged


def str2bool(v: str) -> bool:
    r"""Parse str to bool.

    Args:
      v(str): The string to parse.

    Returns:
      bool: The parsed value.

    """

    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise ValueError("Boolean value expected.")


def str2dict(v: str) -> Dict[str, str]:
    r"""Parse str to dict.

    Args:
      v(str): The string to parse.

    Returns:
      dict: The parsed dict.

    """

    dict = {}
    for token in v.split(","):
        key, value = token.split(":")
        dict[key] = value

    return dict


def str2list(v: str, delim: str = ",") -> List[str]:
    r"""Parse str to list.

    Args:
      v(str): The string to parse.
      delim(str): The delimiter used to split string.

    Returns:
      list: The parsed list.

    """

    return [t.strip() for t in v.split(delim)]


def nullable_float(v: str) -> Optional[float]:
    r"""Parse string to nullable float.

    Args:
      v(str): The string to parse.

    Returns:
      float or None: The parsed value.

    """
    if not v or v.lower() in ["none", "null"]:
        return None
    return float(v)


def nullable_int(v: str) -> Optional[int]:
    r"""Parse string to nullable int.

    Args:
      v(str): The string to parse.

    Returns:
      int or None: The parsed value.

    """
    if not v or v.lower() in ["none", "null"]:
        return None
    return int(v)


def nullable_string(v: str) -> Optional[str]:
    r"""Parse string to nullable string.

    Args:
      v(str): The string to parse.

    Returns:
      str or None: The parsed value.

    """
    if not v or v.lower() in ["none", "null"]:
        return None
    return v


def construct_identifier(
    task_name: str, data_name: str, split_name: str, metric_name: Optional[str] = None
) -> str:
    r"""Construct identifier.

    Args:
      task_name(str): Task name.
      data_name(str): Data set name.
      split_name(str): Split name.
      metric_name(str): Metric name, defaults to None.

    Returns:
      str: The identifier.

    """

    if metric_name:
        return f"{task_name}/{data_name}/{split_name}/{metric_name}"
    else:
        return f"{task_name}/{data_name}/{split_name}"


def random_string(length: int = 5) -> str:
    r"""Generate a random string of fixed length.

    Args:
      length(int): The length of random string, defaults to 5.

    Returns:
      str: The random string.

    """

    letters = string.ascii_lowercase
    return "".join(random.choice(letters) for i in range(length))



# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors, The HuggingFace Inc. team
# and Sen Wu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyTorch optimization for BERT model."""

import logging
from typing import Any, Callable, Iterable, Optional, Tuple

import torch
from torch.optim.optimizer import Optimizer

logger = logging.getLogger(__name__)


class BertAdam(Optimizer):
    r"""Implements BERT version of Adam algorithm with weight decay fix.

    Args:
      params(iterable): Iterable of parameters to optimize or dicts defining
        parameter groups.
      lr(float, optional): Learning rate, defaults to 1e-3.
      betas(Tuple[float, float], optional): Coefficients used for computing running
        averages of gradient and its square, defaults to (0.9, 0.999).
      eps(float, optional): Term added to the denominator to improve numerical
        stability, defaults to 1e-6.
      weight_decay(float, optional): Weight decay (L2 penalty), defaults to 0.01.

    """

    def __init__(
        self,
        params: Iterable[Any],
        lr: float = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-6,
        weight_decay: float = 0.01,
    ) -> None:
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))

        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)  # type: ignore

    def step(self, closure: Optional[Callable] = None) -> Any:
        """Performs a single optimization step.

        Args:
          closure(callable, optional): A closure that reevaluates the model and returns
            the loss, defaults to None.

        """

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:  # type: ignore
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError(
                        "Adam does not support sparse gradients, please consider"
                        "SparseAdam instead"
                    )

                state = self.state[p]  # type: ignore

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                # Decay the first and second moment running average coefficient
                # In-place operations to update the averages at the same time
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                update = exp_avg / (exp_avg_sq.sqrt() + group["eps"])

                # Just adding the square of the weights to the loss function is *not*
                # the correct way of using L2 regularization/weight decay with Adam,
                # since that will interact with the m and v parameters in strange ways.
                #
                # Instead we want to decay the weights in a manner that doesn't interact
                # with the m/v parameters. This is equivalent to adding the square
                # of the weights to the loss with plain (non-momentum) SGD.
                if group["weight_decay"] > 0.0:
                    update += group["weight_decay"] * p.data

                p.data.add_(-group["lr"] * update)

                state["step"] += 1

        return loss



import logging
from functools import partial
from typing import Callable, Dict, List

from numpy import ndarray

from emmental.metrics import METRICS
from emmental.utils.utils import array_to_numpy

logger = logging.getLogger(__name__)


class Scorer(object):
    r"""A class to score tasks.

    Args:
      metrics(list): a list of metric names which provides
        in emmental (e.g., accuracy), defaults to [].
      customize_metric_funcs(dict): a dict of customize metric where key is the metric
        name and value is the metric function which takes gold, preds, probs, uids as
        input, defaults to {}.

    """

    def __init__(
        self, metrics: List[str] = [], customize_metric_funcs: Dict[str, Callable] = {}
    ) -> None:
        self.metrics: Dict[str, Callable] = dict()
        for metric in metrics:
            if metric in METRICS:
                self.metrics[metric] = METRICS[metric]  # type: ignore
            elif metric.startswith("accuracy@"):
                self.metrics[metric] = partial(
                    METRICS["accuracy"], topk=int(metric.split("@")[1])  # type: ignore
                )
            else:
                raise ValueError(f"Unrecognized metric: {metric}")

        self.metrics.update(customize_metric_funcs)

    def score(
        self, golds: ndarray, preds: ndarray, probs: ndarray, uids: List[str] = None
    ) -> Dict[str, float]:
        """Calculate the score.

        Args:
          golds(ndarray): Ground truth values.
          probs(ndarray): Predicted probabilities.
          preds(ndarray): Predicted values.
          uids(list, optional): Unique ids, defaults to None.

        Returns:
          dict: score dict.

        """

        metric_dict = dict()

        for metric_name, metric in self.metrics.items():
            # handle no examples
            if len(golds) == 0:
                metric_dict[metric_name] = float("nan")
                continue

            golds = array_to_numpy(golds)
            preds = array_to_numpy(preds)
            probs = array_to_numpy(probs)

            res = metric(golds, preds, probs, uids)

            if isinstance(res, dict):
                metric_dict.update(res)
            else:
                metric_dict[metric_name] = res

        return metric_dict



from typing import Dict, List, Optional

from numpy import ndarray
from sklearn.metrics import matthews_corrcoef


def matthews_correlation_coefficient_scorer(
    golds: ndarray,
    probs: Optional[ndarray],
    preds: ndarray,
    uids: Optional[List[str]] = None,
) -> Dict[str, float]:
    """Matthews correlation coefficient (MCC).

    Args:
      golds(ndarray): Ground truth values.
      probs(ndarray or None): Predicted probabilities.
      preds(ndarray): Predicted values.
      uids(list, optional): Unique ids, defaults to None.

    Returns:
      dict: Matthews correlation coefficient score.

    """

    return {"matthews_corrcoef": matthews_corrcoef(golds, preds)}



import json
import os
from collections import defaultdict
from typing import Any, Dict, Union

import yaml

from emmental.meta import Meta


class LogWriter(object):
    r"""A class for logging during training process."""

    def __init__(self) -> None:
        """Initialize the log writer.
        """

        self.run_log: defaultdict = defaultdict(list)

    def add_config(self, config: Dict[str, Any]) -> None:
        r"""Log config.

        Args:
          config(dict): The current config.

        """

        self.config = config

    def add_scalar(
        self, name: str, value: Union[float, int], step: Union[float, int]
    ) -> None:
        r"""Log a scalar variable.

        Args:
          name(str): The name of the scalar.
          value(float or int): The value of the scalar.
          step(float or int): The current step.

        """

        self.run_log[name].append((step, value))

    def write_config(self, config_filename: str = "config.yaml") -> None:
        r"""Dump the config to file

        Args:
          config_filename(str, optional): The config filename,
            defaults to "config.yaml".

        """

        config_path = os.path.join(Meta.log_path, config_filename)
        with open(config_path, "w") as yml:
            yaml.dump(self.config, yml, default_flow_style=False, allow_unicode=True)

    def write_log(self, log_filename: str = "log.json") -> None:
        r"""Dump the log to file.

        Args:
          log_filename(str, optional): The log filename, defaults to "log.json".

        """

        log_path = os.path.join(Meta.log_path, log_filename)
        with open(log_path, "w") as f:
            json.dump(self.run_log, f)

    def close(self) -> None:
        r"""Close the log writer."""
        pass



from emmental.contrib.slicing.data import add_slice_labels
from emmental.contrib.slicing.slicing_function import slicing_function
from emmental.contrib.slicing.task import build_slice_tasks

__all__ = ["add_slice_labels", "slicing_function", "build_slice_tasks"]



import math

import torch
import torch.nn as nn
from torch import Tensor


class SparseLinear(nn.Module):
    r"""A sparse linear module.

    Args:
      num_features(int): Size of features.
      num_classes(int): Number of classes.
      bias(bool): Use bias term or not, defaults to False.
      padding_idx(int): padding index, defaults to 0.

    """

    def __init__(
        self,
        num_features: int,
        num_classes: int,
        bias: bool = False,
        padding_idx: int = 0,
    ) -> None:

        super().__init__()

        self.num_features = num_features
        self.num_classes = num_classes
        self.padding_idx = padding_idx

        self.weight = nn.Embedding(
            self.num_features, self.num_classes, padding_idx=self.padding_idx
        )
        if bias:
            self.bias = nn.Parameter(torch.Tensor(self.num_classes))  # type: ignore
        else:
            self.bias = None

        self.reset_parameters()

    def reset_parameters(self) -> None:
        r"""Reinitiate the weight parameters."""

        stdv = 1.0 / math.sqrt(self.num_features)
        self.weight.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)
        if self.padding_idx is not None:
            self.weight.weight.data[self.padding_idx].fill_(0)

    def forward(self, x: Tensor, w: Tensor) -> Tensor:  # type: ignore
        r"""Forward function.

        Args:
          x(Tensor): Feature indices.
          w(Tensor): Feature weights.

        Returns:
          Tensor: Output of linear layer.

        """

        if self.bias is None:
            return (w.unsqueeze(2) * self.weight(x)).sum(dim=1)
        else:
            return (w.unsqueeze(2) * self.weight(x)).sum(dim=1) + self.bias



from typing import Optional

import torch
import torch.nn as nn
from torch import Tensor


class RNN(nn.Module):
    r"""A recurrent neural network module.

    Args:
      num_classes(int): Number of classes.
      emb_size(int): Dimension of embeddings.
      lstm_hidden(int): Size of LSTM hidden layer size.
      num_layers(int): Number of recurrent layers, defaults to 1.
      dropout(float): Dropout parameter of LSTM, defaults to 0.0.
      attention(bool): Use attention or not, defaults to True.
      bidirectional(bool): Use bidirectional LSTM or not, defaults to True.

    """

    def __init__(
        self,
        num_classes: int,
        emb_size: int,
        lstm_hidden: int,
        num_layers: int = 1,
        dropout: float = 0.0,
        attention: bool = True,
        bidirectional: bool = True,
    ) -> None:

        super().__init__()

        self.emb_size = emb_size
        self.lstm_hidden = lstm_hidden
        self.bidirectional = bidirectional
        self.attention = attention
        self.num_layers = num_layers
        self.dropout = dropout
        # Number of dimensions of output. 0 means no final linear layer.
        self.num_classes = num_classes

        if self.num_classes > 0:
            self.final_linear = True
        else:
            self.final_linear = False

        self.drop = nn.Dropout(dropout)

        b = 2 if self.bidirectional else 1

        self.word_lstm = nn.LSTM(
            self.emb_size,
            self.lstm_hidden,
            batch_first=True,
            num_layers=self.num_layers,
            dropout=self.dropout if self.num_layers > 1 else 0.0,
            bidirectional=self.bidirectional,
        )

        if attention:
            self.attn_linear_w_1 = nn.Linear(
                b * lstm_hidden, b * lstm_hidden, bias=True
            )
            self.attn_linear_w_2 = nn.Linear(b * lstm_hidden, 1, bias=False)

        if self.final_linear:
            self.linear = nn.Linear(b * lstm_hidden, num_classes)

    def forward(  # type: ignore
        self, x: Tensor, x_mask: Optional[Tensor] = None
    ) -> Tensor:
        r"""Forward function.

        Args:
          x(Tensor): Input tensor.
          x_mask(Tensor, optional): Input mask tensor, defaults to None.

        Returns:
          Tensor: Output tensor.

        """
        x_emb = self.drop(x)
        output_word, state_word = self.word_lstm(x_emb)
        output_word = self.drop(output_word)
        if self.attention:
            """
            An attention layer where the attention weight is
            a = T' . tanh(Wx + b)
            where x is the input, b is the bias.
            """
            word_squish = torch.tanh(self.attn_linear_w_1(output_word))
            word_attn = self.attn_linear_w_2(word_squish)
            if x_mask is not None:
                word_attn.data.masked_fill_(x_mask.data.unsqueeze(dim=2), float("-inf"))
            word_attn_norm = torch.sigmoid(word_attn.squeeze(2))
            word_attn_vectors = torch.bmm(
                output_word.transpose(1, 2), word_attn_norm.unsqueeze(2)
            ).squeeze(2)
            output = (
                self.linear(word_attn_vectors)
                if self.final_linear
                else word_attn_vectors
            )
        else:
            """
            Mean pooling
            """
            if x_mask is None:
                x_mask = x.new_full(x.size()[:2], fill_value=0, dtype=torch.bool)
            x_lens = x_mask.data.eq(0).long().sum(dim=1)
            weights = (
                output_word.new_ones(output_word.size())
                / x_lens.view(x_lens.size()[0], 1, 1).float()
            )
            weights.data.masked_fill_(x_mask.data.unsqueeze(dim=2), 0.0)
            word_vectors = torch.sum(output_word * weights, dim=1)
            output = self.linear(word_vectors) if self.final_linear else word_vectors

        return output



from emmental.logging.checkpointer import Checkpointer
from emmental.logging.log_writer import LogWriter
from emmental.logging.logging_manager import LoggingManager
from emmental.logging.tensorboard_writer import TensorBoardWriter

__all__ = ["LoggingManager", "LogWriter", "TensorBoardWriter", "Checkpointer"]



from typing import Dict, List, Optional

import numpy as np
from numpy import ndarray

from emmental.utils.utils import prob_to_pred


def precision_scorer(
    golds: ndarray,
    probs: Optional[ndarray],
    preds: ndarray,
    uids: Optional[List[str]] = None,
    pos_label: int = 1,
) -> Dict[str, float]:
    """Precision.

    Args:
      golds(ndarray): Ground truth values.
      probs(ndarray or None): Predicted probabilities.
      preds(ndarray): Predicted values.
      uids(list, optional): Unique ids, defaults to None.
      pos_label(int, optional): The positive class label, defaults to 1.

    Returns:
      dict: Precision.

    """

    # Convert probabilistic label to hard label
    if len(golds.shape) == 2:
        golds = prob_to_pred(golds)

    pred_pos = np.where(preds == pos_label, True, False)
    gt_pos = np.where(golds == pos_label, True, False)
    TP = np.sum(pred_pos * gt_pos)
    FP = np.sum(pred_pos * np.logical_not(gt_pos))

    precision = TP / (TP + FP) if TP + FP > 0 else 0.0

    return {"precision": precision}



from typing import Dict, Iterator, List, Tuple, Union

from torch import Tensor

from emmental.data import EmmentalDataLoader
from emmental.model import EmmentalModel
from emmental.schedulers.scheduler import Scheduler


class SequentialScheduler(Scheduler):
    r"""Generate batch generator from all dataloaders in sequential order for MTL
      training.

    Args:
      fillup(bool): Whether fillup to make all dataloader the same size.

    """

    def __init__(self, fillup: bool = False) -> None:
        super().__init__()

        self.fillup = fillup

    def get_num_batches(self, dataloaders: List[EmmentalDataLoader]) -> int:
        r"""Get total number of batches per epoch.

        Args:
          dataloaders(list): List of dataloaders.

        Returns:
          int: Total number of batches per epoch.

        """

        batch_counts = [len(dataloader) for dataloader in dataloaders]
        if self.fillup:
            batch_counts = [max(batch_counts)] * len(dataloaders)

        for idx in range(len(dataloaders)):
            if dataloaders[idx].n_batches:
                batch_counts[idx] = dataloaders[idx].n_batches

        return sum(batch_counts)

    def get_batches(
        self, dataloaders: List[EmmentalDataLoader], model: EmmentalModel = None
    ) -> Iterator[
        Tuple[
            List[str],
            Dict[str, Union[Tensor, List[str]]],
            Dict[str, Tensor],
            Dict[str, str],
            str,
            str,
        ]
    ]:
        r"""Generate batch generator from all dataloaders in sequential order for
        one epoch.

        Args:
          dataloaders(list): List of dataloaders.
          model(EmmentalModel): The training model, defaults to None.

        Returns:
          genertor: A generator of all batches.

        """

        task_to_label_dicts = [
            dataloader.task_to_label_dict for dataloader in dataloaders
        ]
        uid_names = [dataloader.uid for dataloader in dataloaders]
        data_names = [dataloader.data_name for dataloader in dataloaders]
        splits = [dataloader.split for dataloader in dataloaders]
        data_loaders = [iter(dataloader) for dataloader in dataloaders]

        # Calc the batch size for each dataloader
        batch_counts = [len(dataloader) for dataloader in dataloaders]
        if self.fillup:
            batch_counts = [max(batch_counts)] * len(dataloaders)

        for idx in range(len(dataloaders)):
            if dataloaders[idx].n_batches:
                batch_counts[idx] = dataloaders[idx].n_batches

        for (
            data_loader_idx,
            (task_to_label_dict, data_name, batch_count, split, uid_name),
        ) in enumerate(
            zip(task_to_label_dicts, data_names, batch_counts, splits, uid_names)
        ):
            for batch_idx in range(batch_count):
                try:
                    X_dict, Y_dict = next(data_loaders[data_loader_idx])
                except StopIteration:
                    data_loaders[data_loader_idx] = iter(dataloaders[data_loader_idx])
                    X_dict, Y_dict = next(data_loaders[data_loader_idx])

                yield X_dict[
                    uid_name
                ], X_dict, Y_dict, task_to_label_dict, data_name, split



import copy
import logging
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Tuple

from torch import Tensor
from torch.utils.data import DataLoader, Dataset

from emmental.meta import Meta
from emmental.utils.utils import list_to_tensor, random_string

logger = logging.getLogger(__name__)


class EmmentalDataset(Dataset):
    r"""An advanced dataset class to handle that the input data contains mulitple
    fields and the output data contains multiple label sets.

    Args:
      name(str): The name of the dataset.
      X_dict(dict): The feature dict where key is the feature name and value is the
        feature.
      Y_dict(dict): The label dict where key is the label name and value is
        the label.
      uid(str, optional): The unique id key in the X_dict, defaults to None.

    """

    def __init__(
        self,
        name: str,
        X_dict: Dict[str, Any],
        Y_dict: Dict[str, Tensor],
        uid: Optional[str] = None,
    ) -> None:
        self.name = name
        self.uid = uid
        self.X_dict = X_dict
        self.Y_dict = Y_dict

        if self.uid and self.uid not in self.X_dict:
            raise ValueError(f"Cannot find {self.uid} in X_dict.")

        if self.uid is None:
            self.uid = "_uids_"
            while self.uid in X_dict:
                self.uid = f"_uids_{random_string(3)}_"

            uids = [f"{self.name}_{idx}" for idx in range(self.__len__())]
            self.add_features({f"{self.uid}": uids})

            if Meta.config["meta_config"]["verbose"]:
                logger.info(
                    f"Auto generate uids for dataset {self.name} under {self.uid}."
                )

        for name, label in self.Y_dict.items():
            if not isinstance(label, Tensor):
                raise ValueError(
                    f"Label {name} should be torch.Tensor, not {type(label)}."
                )

    def __getitem__(self, index: int) -> Tuple[Dict[str, Any], Dict[str, Tensor]]:
        r"""Get item by index.

        Args:
          index(index): The index of the item.

        Returns:
          Tuple[Dict[str, Any], Dict[str, Tensor]]: Tuple of x_dict and y_dict

        """

        x_dict = {name: feature[index] for name, feature in self.X_dict.items()}
        y_dict = {name: label[index] for name, label in self.Y_dict.items()}
        return x_dict, y_dict

    def __len__(self) -> int:
        r"""Total number of items in the dataset."""

        try:
            return len(next(iter(self.X_dict.values())))
        except StopIteration:
            return 0

    def _update_dict(self, ori_dict: Dict[str, Any], new_dict: Dict[str, Any]) -> None:
        r"""Update original dict with new dict.

        Args:
          ori_dict(dict): The original dict.
          new_dict(dict): The new dict.

        """

        for key, value in new_dict.items():
            ori_dict[key] = value

    def _remove_key(self, ori_dict: Dict[str, Any], key: str) -> None:
        r"""

        Args:
          ori_dict(dict): The original dict.
          key(str): The key to remove from the original dict.

        """

        if key in ori_dict:
            del ori_dict[key]

    def add_features(self, X_dict: Dict[str, Any]) -> None:
        r"""Add new features into X_dict.

        Args:
          X_dict(dict): The new feature dict to add into the existing feature dict.

        """

        self._update_dict(self.X_dict, X_dict)

    def add_labels(self, Y_dict: Dict[str, Tensor]) -> None:
        r"""Add new labels into Y_dict.

        Args:
          Y_dict(dict): the new label dict to add into the existing label dict

        """

        for name, label in Y_dict.items():
            if not isinstance(label, Tensor):
                raise ValueError(f"Label {name} should be torch.Tensor.")

        self._update_dict(self.Y_dict, Y_dict)

    def remove_feature(self, feature_name: str) -> None:
        r"""Remove one feature from feature dict

        Args:
          feature_name(str): The feature that removes from feature dict.

        """

        self._remove_key(self.X_dict, feature_name)

    def remove_label(self, label_name: str) -> None:
        r"""Remove one label from label dict.

        Args:
          label_name(str): The label that removes from label dict.

        """

        self._remove_key(self.Y_dict, label_name)


def emmental_collate_fn(
    batch: List[Tuple[Dict[str, Any], Dict[str, Tensor]]]
) -> Tuple[Dict[str, Any], Dict[str, Tensor]]:
    r"""Collate function.

    Args:
      batch(Tuple[Dict[str, Any], Dict[str, Tensor]]): The batch to collate.

    Returns:
      Tuple[Dict[str, Any], Dict[str, Tensor]]: The collated batch.

    """
    X_batch: defaultdict = defaultdict(list)
    Y_batch: defaultdict = defaultdict(list)

    for x_dict, y_dict in batch:
        for field_name, value in x_dict.items():
            if isinstance(value, list):
                X_batch[field_name] += value
            else:
                X_batch[field_name].append(value)
        for label_name, value in y_dict.items():
            if isinstance(value, list):
                Y_batch[label_name] += value
            else:
                Y_batch[label_name].append(value)

    field_names = copy.deepcopy(list(X_batch.keys()))

    for field_name in field_names:
        values = X_batch[field_name]
        # Only merge list of tensors
        if isinstance(values[0], Tensor):
            item_tensor, item_mask_tensor = list_to_tensor(
                values,
                min_len=Meta.config["data_config"]["min_data_len"],
                max_len=Meta.config["data_config"]["max_data_len"],
            )
            X_batch[field_name] = item_tensor
            if item_mask_tensor is not None:
                X_batch[f"{field_name}_mask"] = item_mask_tensor

    for label_name, values in Y_batch.items():
        Y_batch[label_name] = list_to_tensor(
            values,
            min_len=Meta.config["data_config"]["min_data_len"],
            max_len=Meta.config["data_config"]["max_data_len"],
        )[0]

    return dict(X_batch), dict(Y_batch)


class EmmentalDataLoader(DataLoader):
    """An advanced dataloader class which contains mapping from task to label (which
    label(s) to use in dataset's Y_dict for this task), and split (which part this
    dataset belongs to) information.

    Args:
      task_to_label_dict(dict): The task to label mapping where key is the task name
        and value is the label(s) for that task and should be the key in Y_dict.
      dataset(EmmentalDataset): The dataset to construct the dataloader
      split(str, optional): The split information, defaults to "train".
      collate_fn(callable, optional): The function that merges a list of samples to
        form a mini-batch, defaults to emmental_collate_fn.
      n_batches(int, optional): Total number of batches.
      **Kwargs: Other arguments of dataloader.

    """

    def __init__(
        self,
        task_to_label_dict: Dict[str, str],
        dataset: EmmentalDataset,
        split: str = "train",
        collate_fn: Callable = emmental_collate_fn,
        n_batches: Optional[int] = None,
        **kwargs: Any,
    ) -> None:

        assert isinstance(dataset, EmmentalDataset)
        super().__init__(dataset, collate_fn=collate_fn, **kwargs)

        self.task_to_label_dict = task_to_label_dict
        self.data_name = dataset.name
        self.uid = dataset.uid
        self.split = split
        self.n_batches = n_batches

        for task_name, label_names in task_to_label_dict.items():
            if not isinstance(label_names, list):
                label_names = [label_names]  # type: ignore
            unrecognized_labels = set(label_names) - set(dataset.Y_dict.keys())
            if len(unrecognized_labels) > 0:
                msg = (
                    f"Unrecognized Label {unrecognized_labels} of Task {task_name} in "
                    f"dataset {dataset.name}"
                )
                logger.warn(msg)
                raise ValueError(msg)



from typing import Dict, List, Optional

from numpy import ndarray
from sklearn.metrics import mean_squared_error


def mean_squared_error_scorer(
    golds: ndarray,
    probs: ndarray,
    preds: Optional[ndarray],
    uids: Optional[List[str]] = None,
) -> Dict[str, float]:
    """Mean squared error regression loss.

    Args:
      golds(ndarray): Ground truth values.
      probs(ndarray): Predicted probabilities.
      preds(ndarray or None): Predicted values.
      uids(list, optional): Unique ids, defaults to None.

    Returns:
      dict: Mean squared error regression loss.

    """

    return {"mean_squared_error": float(mean_squared_error(golds, probs))}



from typing import Any, Dict

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor


class SliceAttentionModule(nn.Module):
    r"""An attention module to leverage all slice representations.

    Args:
      slice_ind_key(str): Slice indicator head key, defaults to "_slice_ind_".
      slice_pred_key(str): Slice prediction head key, defaults to "_slice_pred_",
      slice_pred_feat_key(str): Slice prediction feature key,
        defaults to "_slice_feat_".

    """

    def __init__(
        self,
        slice_ind_key: str = "_slice_ind_",
        slice_pred_key: str = "_slice_pred_",
        slice_pred_feat_key: str = "_slice_feat_",
    ) -> None:

        super().__init__()

        self.slice_ind_key = slice_ind_key
        self.slice_pred_key = slice_pred_key
        self.slice_pred_feat_key = slice_pred_feat_key

    def forward(  # type: ignore
        self, intermediate_output_dict: Dict[str, Any]
    ) -> Tensor:
        r"""Forward function.

        Args:
          intermediate_output_dict(dict): output dict.

        Returns:
          Tensor: output of attention.

        """

        # Collect ordered slice indicator head names
        slice_indicator_names = sorted(
            [
                flow_name
                for flow_name in intermediate_output_dict.keys()
                if self.slice_ind_key in flow_name
            ]
        )
        # Collect ordered slice predictor head names
        slice_predictor_names = sorted(
            [
                flow_name
                for flow_name in intermediate_output_dict.keys()
                if self.slice_pred_key in flow_name
            ]
        )
        # Concat slice indicator predictions
        slice_indicator_predictions = torch.cat(
            [
                F.softmax(intermediate_output_dict[slice_indicator_name][0])[
                    :, 0
                ].unsqueeze(1)
                for slice_indicator_name in slice_indicator_names
            ],
            dim=-1,
        )
        # Concat slice predictor predictions
        slice_predictor_predictions = torch.cat(
            [
                F.softmax(intermediate_output_dict[slice_predictor_name][0])[
                    :, 0
                ].unsqueeze(1)
                for slice_predictor_name in slice_predictor_names
            ],
            dim=-1,
        )
        # Collect ordered slice feature head names
        slice_feature_names = sorted(
            [
                flow_name
                for flow_name in intermediate_output_dict.keys()
                if self.slice_pred_feat_key in flow_name
            ]
        )
        # Concat slice representations
        slice_representations = torch.cat(
            [
                intermediate_output_dict[slice_feature_name][0].unsqueeze(1)
                for slice_feature_name in slice_feature_names
            ],
            dim=1,
        )
        # Attention
        A = (
            F.softmax(slice_indicator_predictions * slice_predictor_predictions, dim=1)
            .unsqueeze(-1)
            .expand([-1, -1, slice_representations.size(-1)])
        )

        reweighted_representation = torch.sum(A * slice_representations, 1)

        return reweighted_representation



import collections
import copy
import logging
from collections import defaultdict
from typing import Dict, List, Optional, Union

import numpy as np
import torch
import torch.optim as optim
from numpy import ndarray
from torch.optim.lr_scheduler import _LRScheduler

from emmental import Meta
from emmental.data import EmmentalDataLoader
from emmental.logging import LoggingManager
from emmental.model import EmmentalModel
from emmental.optimizers.bert_adam import BertAdam
from emmental.schedulers import SCHEDULERS
from emmental.schedulers.scheduler import Scheduler
from emmental.utils.utils import construct_identifier, prob_to_pred

try:
    from IPython import get_ipython

    if "IPKernelApp" not in get_ipython().config:
        raise ImportError("console")
except (AttributeError, ImportError):
    from tqdm import tqdm
else:
    from tqdm import tqdm_notebook as tqdm

logger = logging.getLogger(__name__)


class EmmentalLearner(object):
    r"""A class for emmental multi-task learning.

    Args:
      name(str, optional): Name of the learner, defaults to None.

    """

    def __init__(self, name: Optional[str] = None) -> None:
        self.name = name if name is not None else type(self).__name__

    def _set_logging_manager(self) -> None:
        r"""Set logging manager."""

        self.logging_manager = LoggingManager(self.n_batches_per_epoch)

    def _set_optimizer(self, model: EmmentalModel) -> None:
        r"""Set optimizer for learning process.

        Args:
          model(EmmentalModel): The model to set up the optimizer.

        """
        optimizer_config = Meta.config["learner_config"]["optimizer_config"]
        opt = optimizer_config["optimizer"]

        parameters = filter(lambda p: p.requires_grad, model.parameters())

        optim_dict = {
            # PyTorch optimizer
            "asgd": optim.ASGD,  # type: ignore
            "adadelta": optim.Adadelta,  # type: ignore
            "adagrad": optim.Adagrad,  # type: ignore
            "adam": optim.Adam,  # type: ignore
            "adamw": optim.AdamW,  # type: ignore
            "adamax": optim.Adamax,  # type: ignore
            "lbfgs": optim.LBFGS,  # type: ignore
            "rms_prop": optim.RMSprop,  # type: ignore
            "r_prop": optim.Rprop,  # type: ignore
            "sgd": optim.SGD,  # type: ignore
            "sparse_adam": optim.SparseAdam,  # type: ignore
            # Customize optimizer
            "bert_adam": BertAdam,
        }

        if opt in ["lbfgs", "r_prop", "sparse_adam"]:
            optimizer = optim_dict[opt](
                parameters,
                lr=optimizer_config["lr"],
                **optimizer_config[f"{opt}_config"],
            )
        elif opt in optim_dict.keys():
            optimizer = optim_dict[opt](
                parameters,
                lr=optimizer_config["lr"],
                weight_decay=optimizer_config["l2"],
                **optimizer_config[f"{opt}_config"],
            )
        elif isinstance(opt, optim.Optimizer):  # type: ignore
            optimizer = opt(parameters)
        else:
            raise ValueError(f"Unrecognized optimizer option '{opt}'")

        self.optimizer = optimizer

        if Meta.config["meta_config"]["verbose"]:
            logger.info(f"Using optimizer {self.optimizer}")

    def _set_lr_scheduler(self, model: EmmentalModel) -> None:
        r"""Set learning rate scheduler for learning process.

        Args:
          model(EmmentalModel): The model to set up lr scheduler.

        """

        # Set warmup scheduler
        self._set_warmup_scheduler(model)

        # Set lr scheduler

        lr_scheduler_dict = {
            "exponential": optim.lr_scheduler.ExponentialLR,
            "plateau": optim.lr_scheduler.ReduceLROnPlateau,
            "step": optim.lr_scheduler.StepLR,
            "multi_step": optim.lr_scheduler.MultiStepLR,
            "cyclic": optim.lr_scheduler.CyclicLR,
            "one_cycle": optim.lr_scheduler.OneCycleLR,  # type: ignore
            "cosine_annealing": optim.lr_scheduler.CosineAnnealingLR,
        }

        opt = Meta.config["learner_config"]["lr_scheduler_config"]["lr_scheduler"]
        lr_scheduler_config = Meta.config["learner_config"]["lr_scheduler_config"]

        if opt is None:
            lr_scheduler = None
        elif opt == "linear":
            total_steps = (
                self.n_batches_per_epoch * Meta.config["learner_config"]["n_epochs"]
            )
            linear_decay_func = lambda x: (total_steps - self.warmup_steps - x) / (
                total_steps - self.warmup_steps
            )
            lr_scheduler = optim.lr_scheduler.LambdaLR(
                self.optimizer, linear_decay_func  # type: ignore
            )
        elif opt in ["exponential", "step", "multi_step", "cyclic"]:
            lr_scheduler = lr_scheduler_dict[opt](  # type: ignore
                self.optimizer, **lr_scheduler_config[f"{opt}_config"]
            )
        elif opt == "one_cycle":
            total_steps = (
                self.n_batches_per_epoch * Meta.config["learner_config"]["n_epochs"]
            )
            lr_scheduler = lr_scheduler_dict[opt](  # type: ignore
                self.optimizer,
                total_steps=total_steps,
                epochs=Meta.config["learner_config"]["n_epochs"],
                steps_per_epoch=self.n_batches_per_epoch,
                **lr_scheduler_config[f"{opt}_config"],
            )
        elif opt == "cosine_annealing":
            total_steps = (
                self.n_batches_per_epoch * Meta.config["learner_config"]["n_epochs"]
            )
            lr_scheduler = lr_scheduler_dict[opt](  # type: ignore
                self.optimizer,
                total_steps,
                eta_min=lr_scheduler_config["min_lr"],
                **lr_scheduler_config[f"{opt}_config"],
            )
        elif opt == "plateau":
            plateau_config = copy.deepcopy(lr_scheduler_config["plateau_config"])
            del plateau_config["metric"]
            lr_scheduler = lr_scheduler_dict[opt](
                self.optimizer,
                verbose=Meta.config["meta_config"]["verbose"],
                min_lr=lr_scheduler_config["min_lr"],
                **plateau_config,
            )
        elif isinstance(opt, _LRScheduler):
            lr_scheduler = opt(self.optimizer)  # type: ignore
        else:
            raise ValueError(f"Unrecognized lr scheduler option '{opt}'")

        self.lr_scheduler = lr_scheduler
        self.lr_scheduler_step_unit = Meta.config["learner_config"][
            "lr_scheduler_config"
        ]["lr_scheduler_step_unit"]
        self.lr_scheduler_step_freq = Meta.config["learner_config"][
            "lr_scheduler_config"
        ]["lr_scheduler_step_freq"]

        if Meta.config["meta_config"]["verbose"]:
            logger.info(
                f"Using lr_scheduler {repr(self.lr_scheduler)} with step every "
                f"{self.lr_scheduler_step_freq} {self.lr_scheduler_step_unit}."
            )

    def _set_warmup_scheduler(self, model: EmmentalModel) -> None:
        r"""Set warmup learning rate scheduler for learning process.

        Args:
          model(EmmentalModel): The model to set up warmup scheduler.

        """

        self.warmup_steps = 0
        if Meta.config["learner_config"]["lr_scheduler_config"]["warmup_steps"]:
            warmup_steps = Meta.config["learner_config"]["lr_scheduler_config"][
                "warmup_steps"
            ]
            if warmup_steps < 0:
                raise ValueError(f"warmup_steps much greater or equal than 0.")
            warmup_unit = Meta.config["learner_config"]["lr_scheduler_config"][
                "warmup_unit"
            ]
            if warmup_unit == "epoch":
                self.warmup_steps = int(warmup_steps * self.n_batches_per_epoch)
            elif warmup_unit == "batch":
                self.warmup_steps = int(warmup_steps)
            else:
                raise ValueError(
                    f"warmup_unit must be 'batch' or 'epoch', but {warmup_unit} found."
                )
            linear_warmup_func = lambda x: x / self.warmup_steps
            warmup_scheduler = optim.lr_scheduler.LambdaLR(
                self.optimizer, linear_warmup_func  # type: ignore
            )
            if Meta.config["meta_config"]["verbose"]:
                logger.info(f"Warmup {self.warmup_steps} batchs.")
        elif Meta.config["learner_config"]["lr_scheduler_config"]["warmup_percentage"]:
            warmup_percentage = Meta.config["learner_config"]["lr_scheduler_config"][
                "warmup_percentage"
            ]
            self.warmup_steps = int(
                warmup_percentage
                * Meta.config["learner_config"]["n_epochs"]
                * self.n_batches_per_epoch
            )
            linear_warmup_func = lambda x: x / self.warmup_steps
            warmup_scheduler = optim.lr_scheduler.LambdaLR(
                self.optimizer, linear_warmup_func  # type: ignore
            )
            if Meta.config["meta_config"]["verbose"]:
                logger.info(f"Warmup {self.warmup_steps} batchs.")
        else:
            warmup_scheduler = None

        self.warmup_scheduler = warmup_scheduler

    def _update_lr_scheduler(
        self, model: EmmentalModel, step: int, metric_dict: Dict[str, float]
    ) -> None:
        r"""Update the lr using lr_scheduler with each batch.

        Args:
          model(EmmentalModel): The model to update lr scheduler.
          step(int): The current step.

        """
        cur_lr = self.optimizer.param_groups[0]["lr"]

        if self.warmup_scheduler and step < self.warmup_steps:
            self.warmup_scheduler.step()  # type: ignore
        elif self.lr_scheduler is not None:
            lr_step_cnt = (
                self.lr_scheduler_step_freq
                if self.lr_scheduler_step_unit == "batch"
                else self.lr_scheduler_step_freq * self.n_batches_per_epoch
            )

            if (step + 1) % lr_step_cnt == 0:
                if (
                    Meta.config["learner_config"]["lr_scheduler_config"]["lr_scheduler"]
                    != "plateau"
                ):
                    self.lr_scheduler.step()  # type: ignore
                elif (
                    Meta.config["learner_config"]["lr_scheduler_config"][
                        "plateau_config"
                    ]["metric"]
                    in metric_dict
                ):
                    self.lr_scheduler.step(
                        metric_dict[  # type: ignore
                            Meta.config["learner_config"]["lr_scheduler_config"][
                                "plateau_config"
                            ]["metric"]
                        ]
                    )

            min_lr = Meta.config["learner_config"]["lr_scheduler_config"]["min_lr"]
            if min_lr and self.optimizer.param_groups[0]["lr"] < min_lr:
                self.optimizer.param_groups[0]["lr"] = min_lr

        if (
            Meta.config["learner_config"]["lr_scheduler_config"]["reset_state"]
            and cur_lr != self.optimizer.param_groups[0]["lr"]
        ):
            logger.info("Reset the state of the optimizer.")
            self.optimizer.state = collections.defaultdict(dict)  # Reset state

    def _set_task_scheduler(self) -> None:
        r"""Set task scheduler for learning process"""

        opt = Meta.config["learner_config"]["task_scheduler_config"]["task_scheduler"]

        if opt in ["sequential", "round_robin", "mixed"]:
            self.task_scheduler = SCHEDULERS[opt](  # type: ignore
                **Meta.config["learner_config"]["task_scheduler_config"][
                    f"{opt}_scheduler_config"
                ]
            )
        elif isinstance(opt, Scheduler):
            self.task_scheduler = opt
        else:
            raise ValueError(f"Unrecognized task scheduler option '{opt}'")

    def _evaluate(
        self,
        model: EmmentalModel,
        dataloaders: List[EmmentalDataLoader],
        split: Union[List[str], str],
    ) -> Dict[str, float]:
        r"""Evaluate the model.

        Args:
          model(EmmentalModel): The model to evaluate.
          dataloaders(List[EmmentalDataLoader]): The data to evaluate.
          split(str): The split to evaluate.

        Returns:
          dict: The score dict.

        """

        if not isinstance(split, list):
            valid_split = [split]
        else:
            valid_split = split

        valid_dataloaders = [
            dataloader for dataloader in dataloaders if dataloader.split in valid_split
        ]
        return model.score(valid_dataloaders)

    def _logging(
        self,
        model: EmmentalModel,
        dataloaders: List[EmmentalDataLoader],
        batch_size: int,
    ) -> Dict[str, float]:
        r"""Checking if it's time to evaluting or checkpointing.

        Args:
          model(EmmentalModel): The model to log.
          dataloaders(List[EmmentalDataLoader]): The data to evaluate.
          batch_size(int): Batch size.

        Returns:
          dict: The score dict.

        """

        # Switch to eval mode for evaluation
        model.eval()

        metric_dict = dict()

        self.logging_manager.update(batch_size)

        trigger_evaluation = self.logging_manager.trigger_evaluation()

        # Log the loss and lr
        metric_dict.update(self._aggregate_running_metrics(model, trigger_evaluation))

        # Evaluate the model and log the metric
        if trigger_evaluation:

            # Log task specific metric
            metric_dict.update(
                self._evaluate(
                    model, dataloaders, Meta.config["learner_config"]["valid_split"]
                )
            )

            self.logging_manager.write_log(metric_dict)

            self._reset_losses()

        # Log metric dict every trigger evaluation time or full epoch
        if Meta.config["meta_config"]["verbose"] and (
            trigger_evaluation
            or self.logging_manager.epoch_total == int(self.logging_manager.epoch_total)
        ):
            logger.info(
                f"{self.logging_manager.counter_unit.capitalize()}: "
                f"{self.logging_manager.unit_total:.2f} {metric_dict}"
            )

        # Checkpoint the model
        if self.logging_manager.trigger_checkpointing():
            self.logging_manager.checkpoint_model(
                model, self.optimizer, self.lr_scheduler, metric_dict
            )

            self.logging_manager.write_log(metric_dict)

            self._reset_losses()

        # Switch to train mode
        model.train()

        return metric_dict

    def _aggregate_running_metrics(
        self, model: EmmentalModel, calc_running_scores: bool = False
    ) -> Dict[str, float]:
        r"""Calculate the running overall and task specific metrics.

        Args:
          model(EmmentalModel): The model to evaluate.
          calc_running_scores(bool): Whether to calc running scores

        Returns:
          dict: The score dict.

        """

        metric_dict = dict()

        total_count = 0
        # Log task specific loss
        for identifier in self.running_uids.keys():
            count = len(self.running_uids[identifier])
            if count > 0:
                metric_dict[identifier + "/loss"] = (
                    self.running_losses[identifier] / count
                )
            total_count += count

        # Calculate average micro loss
        if total_count > 0:
            total_loss = sum(self.running_losses.values())
            metric_dict["model/all/train/loss"] = total_loss / total_count

        if calc_running_scores:
            micro_score_dict: Dict[str, List[ndarray]] = defaultdict(list)
            macro_score_dict: Dict[str, List[ndarray]] = defaultdict(list)

            # Calculate training metric
            for identifier in self.running_uids.keys():
                task_name, data_name, split = identifier.split("/")

                metric_score = model.scorers[task_name].score(
                    self.running_golds[identifier],
                    self.running_probs[identifier],
                    prob_to_pred(self.running_probs[identifier]),
                    self.running_uids[identifier],
                )
                for metric_name, metric_value in metric_score.items():
                    metric_dict[f"{identifier}/{metric_name}"] = metric_value

                # Collect average score
                identifier = construct_identifier(
                    task_name, data_name, split, "average"
                )

                metric_dict[identifier] = np.mean(list(metric_score.values()))

                micro_score_dict[split].extend(list(metric_score.values()))
                macro_score_dict[split].append(metric_dict[identifier])

            # Collect split-wise micro/macro average score
            for split in micro_score_dict.keys():
                identifier = construct_identifier(
                    "model", "all", split, "micro_average"
                )
                metric_dict[identifier] = np.mean(micro_score_dict[split])
                identifier = construct_identifier(
                    "model", "all", split, "macro_average"
                )
                metric_dict[identifier] = np.mean(macro_score_dict[split])

        # Log the learning rate
        metric_dict["model/all/train/lr"] = self.optimizer.param_groups[0]["lr"]

        return metric_dict

    def _reset_losses(self) -> None:
        r"""Reset running logs."""

        self.running_uids: Dict[str, List[str]] = defaultdict(list)
        self.running_losses: Dict[str, ndarray] = defaultdict(float)
        self.running_probs: Dict[str, List[ndarray]] = defaultdict(list)
        self.running_golds: Dict[str, List[ndarray]] = defaultdict(list)

    def learn(
        self, model: EmmentalModel, dataloaders: List[EmmentalDataLoader]
    ) -> None:
        r"""The learning procedure of emmental MTL.

        Args:
          model(EmmentalModel): The emmental model that needs to learn.
          dataloaders(List[EmmentalDataLoader]): a list of dataloaders used to
            learn the model.

        """

        # Generate the list of dataloaders for learning process
        train_split = Meta.config["learner_config"]["train_split"]
        if isinstance(train_split, str):
            train_split = [train_split]

        train_dataloaders = [
            dataloader for dataloader in dataloaders if dataloader.split in train_split
        ]

        if not train_dataloaders:
            raise ValueError(
                f"Cannot find the specified train_split "
                f'{Meta.config["learner_config"]["train_split"]} in dataloaders.'
            )

        # Set up task_scheduler
        self._set_task_scheduler()

        # Calculate the total number of batches per epoch
        self.n_batches_per_epoch = self.task_scheduler.get_num_batches(
            train_dataloaders
        )

        # Set up logging manager
        self._set_logging_manager()
        # Set up optimizer
        self._set_optimizer(model)
        # Set up lr_scheduler
        self._set_lr_scheduler(model)

        # Set to training mode
        model.train()

        if Meta.config["meta_config"]["verbose"]:
            logger.info(f"Start learning...")

        self.metrics: Dict[str, float] = dict()
        self._reset_losses()

        for epoch_num in range(Meta.config["learner_config"]["n_epochs"]):
            batches = tqdm(
                enumerate(self.task_scheduler.get_batches(train_dataloaders, model)),
                total=self.n_batches_per_epoch,
                disable=(not Meta.config["meta_config"]["verbose"]),
                desc=f"Epoch {epoch_num}:",
            )

            for batch_num, batch in batches:
                # Covert single batch into a batch list
                if not isinstance(batch, list):
                    batch = [batch]

                total_batch_num = epoch_num * self.n_batches_per_epoch + batch_num
                batch_size = 0

                # Set gradients of all model parameters to zero
                self.optimizer.zero_grad()

                for uids, X_dict, Y_dict, task_to_label_dict, data_name, split in batch:
                    batch_size += len(next(iter(Y_dict.values())))

                    # Perform forward pass and calcualte the loss and count
                    uid_dict, loss_dict, prob_dict, gold_dict = model(
                        uids, X_dict, Y_dict, task_to_label_dict
                    )

                    # Update running loss and count
                    for task_name in uid_dict.keys():
                        identifier = f"{task_name}/{data_name}/{split}"
                        self.running_uids[identifier].extend(uid_dict[task_name])
                        self.running_losses[identifier] += (
                            loss_dict[task_name].item() * len(uid_dict[task_name])
                            if len(loss_dict[task_name].size()) == 0
                            else torch.sum(loss_dict[task_name]).item()
                        )
                        self.running_probs[identifier].extend(prob_dict[task_name])
                        self.running_golds[identifier].extend(gold_dict[task_name])

                    # Skip the backward pass if no loss is calcuated
                    if not loss_dict:
                        continue

                    # Calculate the average loss
                    loss = sum(
                        [
                            model.weights[task_name] * task_loss
                            if len(task_loss.size()) == 0
                            else torch.mean(model.weights[task_name] * task_loss)
                            for task_name, task_loss in loss_dict.items()
                        ]
                    )

                    # Perform backward pass to calculate gradients
                    loss.backward()  # type: ignore

                # Clip gradient norm
                if Meta.config["learner_config"]["optimizer_config"]["grad_clip"]:
                    torch.nn.utils.clip_grad_norm_(
                        model.parameters(),
                        Meta.config["learner_config"]["optimizer_config"]["grad_clip"],
                    )

                # Update the parameters
                self.optimizer.step()

                self.metrics.update(self._logging(model, dataloaders, batch_size))

                batches.set_postfix(self.metrics)

                # Update lr using lr scheduler
                self._update_lr_scheduler(model, total_batch_num, self.metrics)

        model = self.logging_manager.close(model)



from abc import ABC, abstractmethod
from typing import Any, Iterator, List

from emmental.data import EmmentalDataLoader
from emmental.model import EmmentalModel


class Scheduler(ABC):
    r"""Generate batch generator from all dataloaders in designed order for MTL
      training.

    """

    def __init__(self) -> None:

        pass

    def get_num_batches(self, dataloaders: List[EmmentalDataLoader]) -> int:
        r"""Get total number of batches per epoch.

        Args:
          dataloaders(list): List of dataloaders.

        Returns:
          int: Total number of batches per epoch.

        """

        raise NotImplementedError()

    @abstractmethod
    def get_batches(
        self, dataloaders: List[EmmentalDataLoader], model: EmmentalModel = None
    ) -> Iterator[Any]:
        r"""Generate batch generator from all dataloaders in designed order for
        one epoch.

        Args:
          dataloaders(list): List of dataloaders.
          model(EmmentalModel): The training model, defaults to None.

        Returns:
          genertor: A generator of all batches.

        """

        raise NotImplementedError()



from typing import Dict, List, Optional

import numpy as np
from numpy import ndarray
from scipy.stats import spearmanr


def spearman_correlation_scorer(
    golds: ndarray,
    probs: ndarray,
    preds: Optional[ndarray],
    uids: Optional[List[str]] = None,
    return_pvalue: bool = False,
) -> Dict[str, float]:
    r"""Spearman rank-order correlation coefficient and the p-value.

    Args:
      golds(ndarray): Ground truth values.
      probs(ndarray): Predicted probabilities.
      preds(ndarray or None): Predicted values.
      uids(list, optional): Unique ids, defaults to None.
      return_pvalue(bool, optional): Whether return pvalue or not, defaults to False.

    Returns:
      dict: Spearman rank-order correlation coefficient (with pvalue if return_pvalue
      is True).

    """

    probs = np.vstack(probs).squeeze()
    correlation, pvalue = spearmanr(golds, probs)

    if return_pvalue:
        return {"spearman_correlation": correlation, "spearman_pvalue": pvalue}

    return {"spearman_correlation": correlation}



import logging
from typing import Dict, List, Optional

from numpy import ndarray
from sklearn.metrics import roc_auc_score

from emmental.utils.utils import pred_to_prob, prob_to_pred

logger = logging.getLogger(__name__)


def roc_auc_scorer(
    golds: ndarray,
    probs: ndarray,
    preds: Optional[ndarray],
    uids: Optional[List[str]] = None,
) -> Dict[str, float]:
    """ROC AUC.

    Args:
      golds(ndarray): Ground truth values.
      probs(ndarray): Predicted probabilities.
      preds(ndarray or None): Predicted values.
      uids(list, optional): Unique ids, defaults to None.
      pos_label(int, optional): The positive class label, defaults to 1.

    Returns:
      dict: ROC AUC score.

    """

    if len(probs.shape) == 2 and probs.shape[1] == 1:
        probs = probs.reshape(probs.shape[0])

    if len(golds.shape) == 2 and golds.shape[1] == 1:
        golds = golds.reshape(golds.shape[0])

    if len(probs.shape) > 1:
        if len(golds.shape) > 1:
            golds = pred_to_prob(prob_to_pred(golds), n_classes=probs.shape[1])
        else:
            golds = pred_to_prob(golds, n_classes=probs.shape[1])
    else:
        if len(golds.shape) > 1:
            golds = prob_to_pred(golds)

    try:
        roc_auc = roc_auc_score(golds, probs)
    except ValueError:
        logger.warning(
            "Only one class present in golds."
            "ROC AUC score is not defined in that case, set as nan instead."
        )
        roc_auc = float("nan")

    return {"roc_auc": roc_auc}



from typing import Dict, List, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn
from numpy import ndarray
from torch import Tensor


class EmbeddingModule(nn.Module):
    r"""Embedding module.

    Args:
      word_counter(dict): Word count dictionary that contians the frequencies of
        each word, defaults to None.
      max_size(int): Max size of word dictionary, defaults to None.
      word_dim(int): Dimension of embeddings, defaults to 300.
      specials(list): The list of special tokens (e.g., padding or eos) that will
        be prepended to the vocabulary, defaults to [].
      threshold(int): The minimum frequency needed to include a token in the
        vocabulary, defaults to None.
      embedding_file(str): The pretrained embedding file path, defaults to None.
      fix_emb(bool): Whether fix word embeddings or not, defaults to False.

    """

    UNK = "<unk>"
    PAD = "<pad>"

    def __init__(
        self,
        word_counter: Optional[Dict[str, int]] = None,
        max_size: Optional[int] = None,
        word_dim: int = 300,
        specials: List[str] = [],
        threshold: int = 0,
        embedding_file: Optional[str] = None,
        fix_emb: bool = False,
    ) -> None:

        super().__init__()
        assert (
            word_counter is not None or embedding_file is not None
        ), "word_counter and embedding_file are not provided."

        self.word_counter = word_counter
        self.dim = word_dim

        # remove words that occur less than threshold
        if self.word_counter and threshold > 1:
            self.word_counter = dict(
                [(k, v) for k, v in self.word_counter.items() if v >= threshold]
            )

        max_size = None if max_size is None else max_size + len(specials)
        reverse = True

        if embedding_file:
            emb_dim, emb_w2i, emb_wv = self._load_embedding(embedding_file)
            self.dim = emb_dim
            if word_counter is None:
                self.word_counter = emb_w2i
                reverse = False

        self.id2word = sorted(
            self.word_counter, key=lambda k: self.word_counter[k], reverse=reverse
        )

        specials = [self.UNK, self.PAD] + [
            special for special in specials if special not in [self.UNK, self.PAD]
        ]
        # append special tokens and remove duplicate words
        self.id2word = specials + [
            word for word in self.id2word if word not in specials
        ]

        # limit the word list size
        if max_size:
            self.id2word = self.id2word[:max_size]

        self.word2id = dict(
            [(self.id2word[idx], idx) for idx in range(len(self.id2word))]
        )
        self.size = len(self.id2word)

        # Initalize word embeddings
        self.embeddings = nn.Embedding(self.size, self.dim)
        self.embeddings.weight.data.uniform_(-1, 1)
        self.embeddings.weight.data[self.word2id[self.PAD]] = 0.0

        # Update word embedding with pretrained embeddings
        if embedding_file:
            for w, i in emb_w2i.items():
                if w in self.word2id:
                    self.word2id[w]
                    self.embeddings.weight.data[self.word2id[w]].copy_(  # type: ignore
                        torch.from_numpy(emb_wv[emb_w2i[w]])
                    )

        if fix_emb:
            self.embeddings.weight.requires_grad = False

    def _load_embedding(
        self, embedding_file: str
    ) -> Tuple[int, Dict[str, int], List[ndarray]]:
        r"""Load the pre-trained embeddings from file.

        Args:
          embedding_file: The pretrained embedding file path.

        Returns:
          tuple: word embedding dimension, word to index dict, and embedding vectors.

        """

        emb_dim = 0
        emb_w2i: Dict[str, int] = {}
        emb_wv = []

        with open(embedding_file, encoding="utf8") as f:
            for line in f:
                elems = line.split()

                # skip the header
                if len(elems) == 2:
                    continue

                # collect embedding dim
                emb_dim = len(elems) - 1
                # collect word
                token = elems[0]
                # collect word embedding
                if token not in emb_w2i:
                    emb_w2i[token] = len(emb_w2i)
                    emb_wv.append(np.array([float(v) for v in elems[1:]]))
        return emb_dim, emb_w2i, emb_wv

    def forward(self, input: Tensor) -> Tensor:  # type: ignore
        r"""Forward function.

        Args:
          input(Tensor): Input tensor.

        Returns:
          Tensor: Output tensor.

        """
        return self.embeddings(input)



from emmental.metrics.accuracy import accuracy_scorer
from emmental.metrics.accuracy_f1 import accuracy_f1_scorer
from emmental.metrics.fbeta import f1_scorer, fbeta_scorer
from emmental.metrics.matthews_correlation import (
    matthews_correlation_coefficient_scorer,
)
from emmental.metrics.mean_squared_error import mean_squared_error_scorer
from emmental.metrics.pearson_correlation import pearson_correlation_scorer
from emmental.metrics.pearson_spearman import pearson_spearman_scorer
from emmental.metrics.precision import precision_scorer
from emmental.metrics.recall import recall_scorer
from emmental.metrics.roc_auc import roc_auc_scorer
from emmental.metrics.spearman_correlation import spearman_correlation_scorer

METRICS = {
    "accuracy": accuracy_scorer,
    "accuracy_f1": accuracy_f1_scorer,
    "precision": precision_scorer,
    "recall": recall_scorer,
    "f1": f1_scorer,
    "fbeta": fbeta_scorer,
    "matthews_correlation": matthews_correlation_coefficient_scorer,
    "mean_squared_error": mean_squared_error_scorer,
    "pearson_correlation": pearson_correlation_scorer,
    "pearson_spearman": pearson_spearman_scorer,
    "spearman_correlation": spearman_correlation_scorer,
    "roc_auc": roc_auc_scorer,
}

__all__ = [
    "accuracy_scorer",
    "accuracy_f1_scorer",
    "f1_scorer",
    "fbeta_scorer",
    "matthews_correlation_coefficient_scorer",
    "mean_squared_error_scorer",
    "pearson_correlation_scorer",
    "pearson_spearman_scorer",
    "precision_scorer",
    "recall_scorer",
    "roc_auc_scorer",
    "spearman_correlation_scorer",
]



from typing import Any, Dict, Optional

import torch.nn.functional as F
from torch import Tensor


def ce_loss(
    module_name: str,
    intermediate_output_dict: Dict[str, Any],
    Y: Tensor,
    active: Tensor,
    weight: Optional[Tensor] = None,
) -> Tensor:
    """Cross entropy loss.

    Args:
      module_name(str): Module name.
      intermediate_output_dict(dict): output dict.
      Y(Tensor): Gold lables.
      active(Tensor): Active sample index.
      weight(Tensor, optional): Class weights.

    Returns:
      Tensor: Loss.

    """

    return F.cross_entropy(
        intermediate_output_dict[module_name][0][active],
        (Y.view(-1) - 1)[active],
        weight,
    )


def output(module_name: str, intermediate_output_dict: Dict[str, Any]) -> Tensor:
    """Output function.

    Args:
      module_name(str): Module name.
      intermediate_output_dict(dict): output dict.

    Returns:
      Tensor: Output.
    """

    return F.softmax(intermediate_output_dict[module_name][0], dim=1)



import logging
import os
import tempfile
import uuid
from builtins import object
from datetime import datetime
from typing import Any, Dict, Optional, Type

import yaml

from emmental.utils.utils import merge, set_random_seed

MAX_CONFIG_SEARCH_DEPTH = 25  # Max num of parent directories to look for config

logger = logging.getLogger(__name__)


def init(
    log_dir: str = tempfile.gettempdir(),
    log_name: str = "emmental.log",
    use_exact_log_path: bool = False,
    format: str = "[%(asctime)s][%(levelname)s] %(name)s:%(lineno)s - %(message)s",
    level: int = logging.INFO,
    config: Optional[Dict[Any, Any]] = {},
    config_dir: Optional[str] = None,
    config_name: Optional[str] = "emmental-config.yaml",
) -> None:
    r"""Initialize the logging and configuration.

    Args:
      log_dir(str, optional): The directory to store logs in,
        defaults to tempfile.gettempdir().
      log_name(str, optional): The log file name, defaults to "emmental.log".
      use_exact_log_path(bool, optional): Whether to use the exact log directory,
        defaults to False.
      format(str, optional): The logging format string to use,
        defaults to "[%(asctime)s][%(levelname)s] %(name)s:%(lineno)s - %(message)s".
      level(int, optional): The logging level to use, defaults to logging.INFO.
      config(dict, optional): The new configuration, defaults to {}.
      config_dir(str, optional): The path to the config file, defaults to None.
      config_name(str, optional): The config file name,
        defaults to "emmental-config.yaml".

    """

    init_logging(log_dir, log_name, use_exact_log_path, format, level)
    init_config()
    if config or config_dir is not None:
        Meta.update_config(config, config_dir, config_name)

    set_random_seed(Meta.config["meta_config"]["seed"])


def init_config() -> None:
    r"""Load the default configuration."""

    # Load the default setting
    default_config_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "emmental-default-config.yaml"
    )
    with open(default_config_path, "r") as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    logger.info(f"Loading Emmental default config from {default_config_path}.")

    Meta.config = config


def init_logging(
    log_dir: str = tempfile.gettempdir(),
    log_name: str = "emmental.log",
    use_exact_log_path: bool = False,
    format: str = "[%(asctime)s][%(levelname)s] %(name)s:%(lineno)s - %(message)s",
    level: int = logging.INFO,
) -> None:
    r"""Configures logging to output to the provided log_dir.
      Will use a nested directory whose name is the current timestamp.

    Args:
      log_dir(str, optional): The directory to store logs in,
        defaults to tempfile.gettempdir().
      log_name(str, optional): The log file name, defaults to "emmental.log".
      use_exact_log_path(bool, optional): Whether to use the exact log directory,
        defaults to False.
      format(str, optional): The logging format string to use,
        defaults to "[%(asctime)s][%(levelname)s] %(name)s:%(lineno)s - %(message)s".
      level(int, optional): The logging level to use, defaults to logging.INFO.

    """

    if not Meta.log_path:
        if not use_exact_log_path:
            # Generate a new directory using the log_dir, if it doesn't exist
            date = datetime.now().strftime("%Y_%m_%d")
            time = datetime.now().strftime("%H_%M_%S")
            uid = str(uuid.uuid4())[:8]
            log_path = os.path.join(log_dir, date, time, uid)
            while os.path.exists(log_path):
                uid = str(uuid.uuid4())[:8]
                log_path = os.path.join(log_dir, date, time, uid)
        else:
            log_path = log_dir
        os.makedirs(log_path, exist_ok=True)

        # Configure the logger using the provided path
        logging.basicConfig(
            format=format,
            level=level,
            handlers=[
                logging.FileHandler(os.path.join(log_path, log_name)),
                logging.StreamHandler(),
            ],
        )

        # Notify user of log location
        logger.info(f"Setting logging directory to: {log_path}")
        Meta.log_path = log_path
    else:
        logger.info(
            f"Logging was already initialized to use {Meta.log_path}.  "
            "To configure logging manually, call emmental.init_logging before "
            "initialiting Meta."
        )


class Meta(object):
    r"""Singleton-like metadata class for all global variables.
      Adapted from the Unique Design Pattern:
        https://stackoverflow.com/questions/1318406/why-is-the-borg-pattern-better-than-the-singleton-pattern-in-python

    """

    log_path: Optional[str] = None
    config: Optional[Dict[Any, Any]] = None

    @classmethod
    def init(cls) -> Type["Meta"]:
        """ """
        if not Meta.log_path:
            init_logging()

        if not Meta.config:
            init_config()

        return cls

    @staticmethod
    def update_config(
        config: Optional[Dict[Any, Any]] = {},
        path: Optional[str] = None,
        filename: Optional[str] = "emmental-config.yaml",
    ) -> None:
        r"""Update the configuration with the configs in root of project and
          its parents.

        Note: There are two ways to update the config:
            (1) uses a config dict to update to config
            (2) uses path and filename to load yaml file to update config

        Args:
          config(dict, optional): The new configuration, defaults to {}.
          path(str, optional): The path to the config file, defaults to os.getcwd().
          filename(str, optional): The config file name,
            defaults to "emmental-config.yaml".

        """

        if config != {}:
            Meta.config = merge(Meta.config, config, specical_keys="checkpoint_metric")
            logger.info("Updating Emmental config from user provided config.")

        if path is not None:
            tries = 0
            current_dir = path
            while current_dir and tries < MAX_CONFIG_SEARCH_DEPTH:
                potential_path = os.path.join(current_dir, filename)
                if os.path.exists(potential_path):
                    with open(potential_path, "r") as f:
                        Meta.config = merge(
                            Meta.config,
                            yaml.load(f, Loader=yaml.FullLoader),
                            specical_keys="checkpoint_metric",
                        )
                    logger.info(f"Updating Emmental config from {potential_path}.")
                    break

                new_dir = os.path.split(current_dir)[0]
                if current_dir == new_dir:
                    logger.info("Unable to find config file. Using defaults.")
                    break
                current_dir = new_dir
                tries += 1

    @staticmethod
    def reset() -> None:
        r"""Clears shared variables of shared, global singleton."""

        Meta.log_path = None
        Meta.config = None



import logging
from typing import Any, Callable, Dict, List

import numpy as np
import torch
from torch import Tensor

from emmental.contrib.slicing.slicing_function import slicing_function
from emmental.data import EmmentalDataLoader
from emmental.task import EmmentalTask

logger = logging.getLogger(__name__)


def add_slice_labels(
    task: EmmentalTask,
    dataloaders: List[EmmentalDataLoader],
    slice_func_dict: Dict[str, Callable],
    split: str = "train",
) -> Dict[str, Tensor]:
    r"""A function to extend dataloader by adding slice indicator and predictor
    labels.

    Args:
      task(EmmentalTask): Task to add slices.
      dataloaders(List[EmmentalDataLoader]): List of dataloaders to train on the task.
      slice_func_dict(dict): Slicing functions.
      split(str): Split to use, defaults to "train".

    Returns:
      dict: slice data class distribution.

    """

    # Calculate class balance
    slice_distribution = {}

    # Add base slice if needed
    if "base" not in slice_func_dict.keys():
        slice_func_dict["base"] = base_slice

    for dataloader in dataloaders:
        labels = dataloader.dataset.Y_dict[  # type: ignore
            dataloader.task_to_label_dict[task.name]
        ]
        for slice_name, slice_func in slice_func_dict.items():
            indicators = slice_func(dataloader.dataset)
            slice_ind_name = f"{task.name}_slice:ind_{slice_name}"
            slice_pred_name = f"{task.name}_slice:pred_{slice_name}"

            pred_labels = indicators * labels
            ind_labels = indicators
            ind_labels[ind_labels == 0] = 2

            if dataloader.split == split and slice_name != "base":
                ind_classes, ind_counts = np.unique(
                    ind_labels.numpy(), return_counts=True
                )
                if ind_classes.shape[0] == 2:
                    slice_distribution[slice_ind_name] = torch.Tensor(  # type: ignore
                        np.sum(ind_counts) / ind_counts / ind_classes.shape[0]
                    )
                pred_classes, pred_counts = np.unique(
                    pred_labels.numpy(), return_counts=True
                )
                if (pred_classes[0] == 0 and pred_classes.shape[0] == 3) or (
                    pred_classes[0] == 1 and pred_classes.shape[0] == 2
                ):
                    if pred_classes[0] == 0:
                        slice_distribution[
                            slice_pred_name
                        ] = torch.Tensor(  # type: ignore
                            1 - pred_counts[1:] / np.sum(pred_counts[1:])
                        )
                    else:
                        slice_distribution[
                            slice_pred_name
                        ] = torch.Tensor(  # type: ignore
                            1 - pred_counts / np.sum(pred_counts)
                        )

            # Update slice indicator and predictor labels
            dataloader.dataset.Y_dict.update(  # type: ignore
                {slice_ind_name: ind_labels, slice_pred_name: pred_labels}
            )
            # Update dataloader task_to_label_dict
            dataloader.task_to_label_dict.update(
                {slice_ind_name: slice_ind_name, slice_pred_name: slice_pred_name}
            )
        msg = (
            f"Loaded slice labels for task {task.name}, slice {slice_name}, "
            f"split {dataloader.split}."
        )
        logger.info(msg)

    return slice_distribution


@slicing_function()
def base_slice(example: Any) -> bool:
    r"""Base slice which always to return True.

    Args:
      example(Any): Sample to check if it's in the slice or not.

    Returns:
      bool: True

    """
    return True



import argparse
from argparse import ArgumentParser, Namespace
from typing import Any, Dict, Optional

from emmental.utils.utils import (
    nullable_float,
    nullable_int,
    nullable_string,
    str2bool,
    str2dict,
)


def parse_args(parser: Optional[ArgumentParser] = None) -> ArgumentParser:
    r"""Parse the configuration from command line.

    Args:
      parser(ArgumentParser): The exterenl argument parser object, defaults to None.

    Returns:
      ArgumentParser: The updated argument parser object.

    """

    if parser is None:
        parser = argparse.ArgumentParser(
            "Emmental configuration",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

    # Load meta configuration
    meta_config = parser.add_argument_group("Meta configuration")

    meta_config.add_argument(
        "--seed",
        type=nullable_int,
        default=None,
        help="Random seed for all numpy/torch/cuda operations in model and learning",
    )

    meta_config.add_argument(
        "--verbose",
        type=str2bool,
        default=True,
        help="Whether to print the log information",
    )

    meta_config.add_argument(
        "--log_path", type=str, default="logs", help="Directory to save running log"
    )

    meta_config.add_argument(
        "--use_exact_log_path",
        type=str2bool,
        default=False,
        help="Whether to use the exact log directory",
    )

    # Load data configuration
    data_config = parser.add_argument_group("Data configuration")

    data_config.add_argument(
        "--min_data_len", type=int, default=0, help="Minimal data length"
    )

    data_config.add_argument(
        "--max_data_len",
        type=int,
        default=0,
        help="Maximal data length (0 for no max_len)",
    )

    # Load model configuration
    model_config = parser.add_argument_group("Model configuration")

    model_config.add_argument(
        "--model_path",
        type=nullable_string,
        default=None,
        help="Path to pretrained model",
    )

    model_config.add_argument(
        "--device",
        type=int,
        default=0,
        help="Which device to use (-1 for cpu or gpu id (e.g., 0 for cuda:0))",
    )

    model_config.add_argument(
        "--dataparallel",
        type=str2bool,
        default=True,
        help="Whether to use dataparallel or not",
    )

    # Learning configuration
    learner_config = parser.add_argument_group("Learning configuration")

    learner_config.add_argument(
        "--fp16",
        type=str2bool,
        default=False,
        help="Whether to use half precision to train",
    )

    learner_config.add_argument(
        "--n_epochs", type=int, default=1, help="Total number of learning epochs"
    )

    learner_config.add_argument(
        "--train_split",
        nargs="+",
        type=str,
        default=["train"],
        help="The split for training",
    )

    learner_config.add_argument(
        "--valid_split",
        nargs="+",
        type=str,
        default=["valid"],
        help="The split for validation",
    )

    learner_config.add_argument(
        "--test_split",
        nargs="+",
        type=str,
        default=["test"],
        help="The split for testing",
    )

    learner_config.add_argument(
        "--ignore_index",
        type=nullable_int,
        default=None,
        help="The ignore index, uses for masking samples",
    )

    # Optimizer configuration
    optimizer_config = parser.add_argument_group("Optimizer configuration")

    optimizer_config.add_argument(
        "--optimizer",
        type=nullable_string,
        default="adam",
        choices=[
            "asgd",
            "adadelta",
            "adagrad",
            "adam",
            "adamw",
            "adamax",
            "lbfgs",
            "rms_prop",
            "r_prop",
            "sgd",
            "sparse_adam",
            "bert_adam",
            None,
        ],
        help="The optimizer to use",
    )

    optimizer_config.add_argument("--lr", type=float, default=1e-3, help="Learing rate")

    optimizer_config.add_argument(
        "--l2", type=float, default=0.0, help="l2 regularization"
    )

    optimizer_config.add_argument(
        "--grad_clip", type=nullable_float, default=None, help="Gradient clipping"
    )

    # ASGD config
    optimizer_config.add_argument(
        "--asgd_lambd", type=float, default=0.0001, help="ASGD lambd"
    )

    optimizer_config.add_argument(
        "--asgd_alpha", type=float, default=0.75, help="ASGD alpha"
    )

    optimizer_config.add_argument(
        "--asgd_t0", type=float, default=1000000.0, help="ASGD t0"
    )

    # Adadelta config
    optimizer_config.add_argument(
        "--adadelta_rho", type=float, default=0.9, help="Adadelta rho"
    )

    optimizer_config.add_argument(
        "--adadelta_eps", type=float, default=0.000001, help="Adadelta eps"
    )

    # Adagrad config
    optimizer_config.add_argument(
        "--adagrad_lr_decay", type=float, default=0, help="Adagrad lr_decay"
    )

    optimizer_config.add_argument(
        "--adagrad_initial_accumulator_value",
        type=float,
        default=0,
        help="Adagrad initial accumulator value",
    )

    optimizer_config.add_argument(
        "--adagrad_eps", type=float, default=0.0000000001, help="Adagrad eps"
    )

    # Adam config
    optimizer_config.add_argument(
        "--adam_betas", nargs="+", type=float, default=(0.9, 0.999), help="Adam betas"
    )

    optimizer_config.add_argument(
        "--adam_eps", type=float, default=1e-8, help="Adam eps"
    )

    optimizer_config.add_argument(
        "--adam_amsgrad",
        type=str2bool,
        default=False,
        help="Whether to use the AMSGrad variant of adam",
    )

    # AdamW config
    optimizer_config.add_argument(
        "--adamw_betas", nargs="+", type=float, default=(0.9, 0.999), help="AdamW betas"
    )

    optimizer_config.add_argument(
        "--adamw_eps", type=float, default=1e-8, help="AdamW eps"
    )

    optimizer_config.add_argument(
        "--adamw_amsgrad",
        type=str2bool,
        default=False,
        help="Whether to use the AMSGrad variant of AdamW",
    )

    # Adamax config
    optimizer_config.add_argument(
        "--adamax_betas",
        nargs="+",
        type=float,
        default=(0.9, 0.999),
        help="Adamax betas",
    )

    optimizer_config.add_argument(
        "--adamax_eps", type=float, default=1e-8, help="Adamax eps"
    )

    # LBFGS config
    optimizer_config.add_argument(
        "--lbfgs_max_iter", type=int, default=20, help="LBFGS max iter"
    )

    optimizer_config.add_argument(
        "--lbfgs_max_eval", type=nullable_int, default=None, help="LBFGS max eval"
    )

    optimizer_config.add_argument(
        "--lbfgs_tolerance_grad", type=float, default=1e-07, help="LBFGS tolerance grad"
    )

    optimizer_config.add_argument(
        "--lbfgs_tolerance_change",
        type=float,
        default=1e-09,
        help="LBFGS tolerance change",
    )

    optimizer_config.add_argument(
        "--lbfgs_history_size", type=int, default=100, help="LBFGS history size"
    )

    optimizer_config.add_argument(
        "--lbfgs_line_search_fn",
        type=nullable_string,
        default=None,
        help="LBFGS line search fn",
    )

    # RMSprop config
    optimizer_config.add_argument(
        "--rms_prop_alpha", type=float, default=0.99, help="RMSprop alpha"
    )

    optimizer_config.add_argument(
        "--rms_prop_eps", type=float, default=1e-08, help="RMSprop eps"
    )

    optimizer_config.add_argument(
        "--rms_prop_momentum", type=float, default=0, help="RMSprop momentum"
    )

    optimizer_config.add_argument(
        "--rms_prop_centered", type=str2bool, default=False, help="RMSprop centered"
    )

    # Rprop config
    optimizer_config.add_argument(
        "--r_prop_etas", nargs="+", type=float, default=(0.5, 1.2), help="Rprop etas"
    )

    optimizer_config.add_argument(
        "--r_prop_step_sizes",
        nargs="+",
        type=float,
        default=(1e-06, 50),
        help="Rprop step sizes",
    )

    # SGD config
    optimizer_config.add_argument(
        "--sgd_momentum", type=float, default=0, help="SGD momentum"
    )

    optimizer_config.add_argument(
        "--sgd_dampening", type=float, default=0, help="SGD dampening"
    )

    optimizer_config.add_argument(
        "--sgd_nesterov", type=str2bool, default=False, help="SGD nesterov"
    )

    # SparseAdam config
    optimizer_config.add_argument(
        "--sparse_adam_betas",
        nargs="+",
        type=float,
        default=(0.9, 0.999),
        help="SparseAdam betas",
    )

    optimizer_config.add_argument(
        "--sparse_adam_eps", type=float, default=1e-08, help="SparseAdam eps"
    )

    # BertAdam config
    optimizer_config.add_argument(
        "--bert_adam_betas",
        nargs="+",
        type=float,
        default=(0.9, 0.999),
        help="BertAdam betas",
    )

    optimizer_config.add_argument(
        "--bert_adam_eps", type=float, default=1e-08, help="BertAdam eps"
    )

    # Scheduler configuration
    scheduler_config = parser.add_argument_group("Scheduler configuration")

    scheduler_config.add_argument(
        "--lr_scheduler",
        type=nullable_string,
        default=None,
        choices=[
            "linear",
            "exponential",
            "plateau",
            "step",
            "multi_step",
            "cyclic",
            "one_cycle",
            "cosine_annealing",
        ],
        help="Learning rate scheduler",
    )

    scheduler_config.add_argument(
        "--lr_scheduler_step_unit",
        type=str,
        default="batch",
        choices=["batch", "epoch"],
        help="Learning rate scheduler step unit",
    )

    scheduler_config.add_argument(
        "--lr_scheduler_step_freq",
        type=int,
        default=1,
        help="Learning rate scheduler step freq",
    )

    scheduler_config.add_argument(
        "--warmup_steps", type=float, default=None, help="Warm up steps"
    )

    scheduler_config.add_argument(
        "--warmup_unit",
        type=str,
        default="batch",
        choices=["batch", "epoch"],
        help="Warm up unit",
    )

    scheduler_config.add_argument(
        "--warmup_percentage", type=float, default=None, help="Warm up percentage"
    )

    scheduler_config.add_argument(
        "--min_lr", type=float, default=0.0, help="Minimum learning rate"
    )

    scheduler_config.add_argument(
        "--reset_state",
        type=str2bool,
        default=False,
        help="Whether reset the state of the optimizer when lr changes",
    )

    scheduler_config.add_argument(
        "--exponential_lr_scheduler_gamma",
        type=float,
        default=0.9,
        help="Gamma for exponential lr scheduler",
    )

    # ReduceLROnPlateau lr scheduler config
    scheduler_config.add_argument(
        "--plateau_lr_scheduler_metric",
        type=str,
        default="model/train/all/loss",
        help="Metric of plateau lr scheduler",
    )

    scheduler_config.add_argument(
        "--plateau_lr_scheduler_mode",
        type=str,
        default="min",
        choices=["min", "max"],
        help="Mode of plateau lr scheduler",
    )

    scheduler_config.add_argument(
        "--plateau_lr_scheduler_factor",
        type=float,
        default=0.1,
        help="Factor of plateau lr scheduler",
    )

    scheduler_config.add_argument(
        "--plateau_lr_scheduler_patience",
        type=int,
        default=10,
        help="Patience for plateau lr scheduler",
    )

    scheduler_config.add_argument(
        "--plateau_lr_scheduler_threshold",
        type=float,
        default=0.0001,
        help="Threshold of plateau lr scheduler",
    )

    scheduler_config.add_argument(
        "--plateau_lr_scheduler_threshold_mode",
        type=str,
        default="rel",
        choices=["rel", "abs"],
        help="Threshold mode of plateau lr scheduler",
    )

    scheduler_config.add_argument(
        "--plateau_lr_scheduler_cooldown",
        type=int,
        default=0,
        help="Cooldown of plateau lr scheduler",
    )

    scheduler_config.add_argument(
        "--plateau_lr_scheduler_eps",
        type=float,
        default=0.00000001,
        help="Eps of plateau lr scheduler",
    )

    # Step lr scheduler config
    scheduler_config.add_argument(
        "--step_lr_scheduler_step_size",
        type=int,
        default=1,
        help="Period of learning rate decay",
    )

    scheduler_config.add_argument(
        "--step_lr_scheduler_gamma",
        type=float,
        default=0.1,
        help="Multiplicative factor of learning rate decay",
    )

    scheduler_config.add_argument(
        "--step_lr_scheduler_last_epoch",
        type=int,
        default=-1,
        help="The index of last epoch",
    )

    scheduler_config.add_argument(
        "--multi_step_lr_scheduler_milestones",
        nargs="+",
        type=int,
        default=[1000],
        help="List of epoch indices. Must be increasing.",
    )

    scheduler_config.add_argument(
        "--multi_step_lr_scheduler_gamma",
        type=float,
        default=0.1,
        help="Multiplicative factor of learning rate decay",
    )

    scheduler_config.add_argument(
        "--multi_step_lr_scheduler_last_epoch",
        type=int,
        default=-1,
        help="The index of last epoch",
    )

    # Cyclic lr scheduler config
    scheduler_config.add_argument(
        "--cyclic_lr_scheduler_base_lr",
        nargs="+",
        type=float,
        default=0.001,
        help="Base lr of cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--cyclic_lr_scheduler_max_lr",
        nargs="+",
        type=float,
        default=0.1,
        help="Max lr of cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--cyclic_lr_scheduler_step_size_up",
        type=int,
        default=2000,
        help="Step size up of cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--cyclic_lr_scheduler_step_size_down",
        type=nullable_int,
        default=None,
        help="Step size down of cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--cyclic_lr_scheduler_mode",
        type=nullable_string,
        default="triangular",
        help="Mode of cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--cyclic_lr_scheduler_gamma",
        type=float,
        default=1.0,
        help="Gamma of cyclic lr scheduler",
    )

    # TODO: support cyclic_lr_scheduler_scale_fn

    scheduler_config.add_argument(
        "--cyclic_lr_scheduler_scale_mode",
        type=str,
        default="cycle",
        choices=["cycle", "iterations"],
        help="Scale mode of cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--cyclic_lr_scheduler_cycle_momentum",
        type=str2bool,
        default=True,
        help="Cycle momentum of cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--cyclic_lr_scheduler_base_momentum",
        nargs="+",
        type=float,
        default=0.8,
        help="Base momentum of cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--cyclic_lr_scheduler_max_momentum",
        nargs="+",
        type=float,
        default=0.9,
        help="Max momentum of cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--cyclic_lr_scheduler_last_epoch",
        type=int,
        default=-1,
        help="Last epoch of cyclic lr scheduler",
    )

    # One cycle lr scheduler config
    scheduler_config.add_argument(
        "--one_cycle_lr_scheduler_max_lr",
        nargs="+",
        type=float,
        default=0.1,
        help="Max lr of one cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--one_cycle_lr_scheduler_pct_start",
        type=float,
        default=0.3,
        help="Percentage start of one cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--one_cycle_lr_scheduler_anneal_strategy",
        type=str,
        default="cos",
        choices=["cos", "linear"],
        help="Anneal strategyr of one cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--one_cycle_lr_scheduler_cycle_momentum",
        type=str2bool,
        default=True,
        help="Cycle momentum of one cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--one_cycle_lr_scheduler_base_momentum",
        nargs="+",
        type=float,
        default=0.85,
        help="Base momentum of one cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--one_cycle_lr_scheduler_max_momentum",
        nargs="+",
        type=float,
        default=0.95,
        help="Max momentum of one cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--one_cycle_lr_scheduler_div_factor",
        type=float,
        default=25,
        help="Div factor of one cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--one_cycle_lr_scheduler_final_div_factor",
        type=float,
        default=1e4,
        help="Final div factor of one cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--one_cycle_lr_scheduler_last_epoch",
        type=int,
        default=-1,
        help="Last epoch of one cyclic lr scheduler",
    )

    scheduler_config.add_argument(
        "--cosine_annealing_lr_scheduler_last_epoch",
        type=int,
        default=-1,
        help="The index of last epoch",
    )

    scheduler_config.add_argument(
        "--task_scheduler",
        type=str,
        default="round_robin",
        # choices=["sequential", "round_robin", "mixed"],
        help="Task scheduler",
    )

    scheduler_config.add_argument(
        "--sequential_scheduler_fillup",
        type=str2bool,
        default=False,
        help="Whether fillup in sequential scheduler",
    )

    scheduler_config.add_argument(
        "--round_robin_scheduler_fillup",
        type=str2bool,
        default=False,
        help="whether fillup in round robin scheduler",
    )

    scheduler_config.add_argument(
        "--mixed_scheduler_fillup",
        type=str2bool,
        default=False,
        help="whether fillup in mixed scheduler scheduler",
    )

    # Logging configuration
    logging_config = parser.add_argument_group("Logging configuration")

    logging_config.add_argument(
        "--counter_unit",
        type=str,
        default="epoch",
        choices=["epoch", "batch"],
        help="Logging unit (epoch, batch)",
    )

    logging_config.add_argument(
        "--evaluation_freq", type=float, default=1, help="Logging evaluation frequency"
    )

    logging_config.add_argument(
        "--writer",
        type=str,
        default="tensorboard",
        choices=["json", "tensorboard"],
        help="The writer format (json, tensorboard)",
    )

    logging_config.add_argument(
        "--checkpointing",
        type=str2bool,
        default=False,
        help="Whether to checkpoint the model",
    )

    logging_config.add_argument(
        "--checkpoint_path", type=str, default=None, help="Checkpointing path"
    )

    logging_config.add_argument(
        "--checkpoint_freq",
        type=int,
        default=1,
        help="Checkpointing every k logging time",
    )

    logging_config.add_argument(
        "--checkpoint_metric",
        type=str2dict,
        default={"model/train/all/loss": "min"},
        help=(
            "Checkpointing metric (metric_name:mode), "
            "e.g., `model/train/all/loss:min`"
        ),
    )

    logging_config.add_argument(
        "--checkpoint_task_metrics",
        type=str2dict,
        default=None,
        help=(
            "Task specific checkpointing metric "
            "(metric_name1:mode1,metric_name2:mode2)"
        ),
    )

    logging_config.add_argument(
        "--checkpoint_runway",
        type=float,
        default=0,
        help="Checkpointing runway (no checkpointing before k checkpointing unit)",
    )

    logging_config.add_argument(
        "--checkpoint_all",
        type=str2bool,
        default=False,
        help="Whether to checkpoint all checkpoints",
    )

    logging_config.add_argument(
        "--clear_intermediate_checkpoints",
        type=str2bool,
        default=True,
        help="Whether to clear intermediate checkpoints",
    )

    logging_config.add_argument(
        "--clear_all_checkpoints",
        type=str2bool,
        default=False,
        help="Whether to clear all checkpoints",
    )

    return parser


def parse_args_to_config(args: Namespace) -> Dict[str, Any]:
    r"""Parse the arguments to config dict

    Args:
      args(Namespace): The parsed namespace from argument parser.

    Returns:
      dict: The config dict.

    """
    config = {
        "meta_config": {
            "seed": args.seed,
            "verbose": args.verbose,
            "log_path": args.log_path,
            "use_exact_log_path": args.use_exact_log_path,
        },
        "data_config": {
            "min_data_len": args.min_data_len,
            "max_data_len": args.max_data_len,
        },
        "model_config": {
            "model_path": args.model_path,
            "device": args.device,
            "dataparallel": args.dataparallel,
        },
        "learner_config": {
            "fp16": args.fp16,
            "n_epochs": args.n_epochs,
            "train_split": args.train_split,
            "valid_split": args.valid_split,
            "test_split": args.test_split,
            "ignore_index": args.ignore_index,
            "optimizer_config": {
                "optimizer": args.optimizer,
                "lr": args.lr,
                "l2": args.l2,
                "grad_clip": args.grad_clip,
                "asgd_config": {
                    "lambd": args.asgd_lambd,
                    "alpha": args.asgd_alpha,
                    "t0": args.asgd_t0,
                },
                "adadelta_config": {"rho": args.adadelta_rho, "eps": args.adadelta_eps},
                "adagrad_config": {
                    "lr_decay": args.adagrad_lr_decay,
                    "initial_accumulator_value": args.adagrad_initial_accumulator_value,
                    "eps": args.adagrad_eps,
                },
                "adam_config": {
                    "betas": args.adam_betas,
                    "amsgrad": args.adam_amsgrad,
                    "eps": args.adam_eps,
                },
                "adamw_config": {
                    "betas": args.adamw_betas,
                    "amsgrad": args.adamw_amsgrad,
                    "eps": args.adamw_eps,
                },
                "adamax_config": {"betas": args.adamax_betas, "eps": args.adamax_eps},
                "lbfgs_config": {
                    "max_iter": args.lbfgs_max_iter,
                    "max_eval": args.lbfgs_max_eval,
                    "tolerance_grad": args.lbfgs_tolerance_grad,
                    "tolerance_change": args.lbfgs_tolerance_change,
                    "history_size": args.lbfgs_history_size,
                    "line_search_fn": args.lbfgs_line_search_fn,
                },
                "rms_prop_config": {
                    "alpha": args.rms_prop_alpha,
                    "eps": args.rms_prop_eps,
                    "momentum": args.rms_prop_momentum,
                    "centered": args.rms_prop_centered,
                },
                "r_prop_config": {
                    "etas": args.r_prop_etas,
                    "step_sizes": args.r_prop_step_sizes,
                },
                "sgd_config": {
                    "momentum": args.sgd_momentum,
                    "dampening": args.sgd_dampening,
                    "nesterov": args.sgd_nesterov,
                },
                "sparse_adam_config": {
                    "betas": args.sparse_adam_betas,
                    "eps": args.sparse_adam_eps,
                },
                "bert_adam_config": {
                    "betas": args.bert_adam_betas,
                    "eps": args.bert_adam_eps,
                },
            },
            "lr_scheduler_config": {
                "lr_scheduler": args.lr_scheduler,
                "lr_scheduler_step_unit": args.lr_scheduler_step_unit,
                "lr_scheduler_step_freq": args.lr_scheduler_step_freq,
                "warmup_steps": args.warmup_steps,
                "warmup_unit": args.warmup_unit,
                "warmup_percentage": args.warmup_percentage,
                "min_lr": args.min_lr,
                "reset_state": args.reset_state,
                "exponential_config": {"gamma": args.exponential_lr_scheduler_gamma},
                "plateau_config": {
                    "metric": args.plateau_lr_scheduler_metric,
                    "mode": args.plateau_lr_scheduler_mode,
                    "factor": args.plateau_lr_scheduler_factor,
                    "patience": args.plateau_lr_scheduler_patience,
                    "threshold": args.plateau_lr_scheduler_threshold,
                    "threshold_mode": args.plateau_lr_scheduler_threshold_mode,
                    "cooldown": args.plateau_lr_scheduler_cooldown,
                    "eps": args.plateau_lr_scheduler_eps,
                },
                "step_config": {
                    "step_size": args.step_lr_scheduler_step_size,
                    "gamma": args.step_lr_scheduler_gamma,
                    "last_epoch": args.step_lr_scheduler_last_epoch,
                },
                "multi_step_config": {
                    "milestones": args.multi_step_lr_scheduler_milestones,
                    "gamma": args.multi_step_lr_scheduler_gamma,
                    "last_epoch": args.multi_step_lr_scheduler_last_epoch,
                },
                "cyclic_config": {
                    "base_lr": args.cyclic_lr_scheduler_base_lr,
                    "max_lr": args.cyclic_lr_scheduler_max_lr,
                    "step_size_up": args.cyclic_lr_scheduler_step_size_up,
                    "step_size_down": args.cyclic_lr_scheduler_step_size_down,
                    "mode": args.cyclic_lr_scheduler_mode,
                    "gamma": args.cyclic_lr_scheduler_gamma,
                    "scale_fn": None,
                    "scale_mode": args.cyclic_lr_scheduler_scale_mode,
                    "cycle_momentum": args.cyclic_lr_scheduler_cycle_momentum,
                    "base_momentum": args.cyclic_lr_scheduler_base_momentum,
                    "max_momentum": args.cyclic_lr_scheduler_max_momentum,
                    "last_epoch": args.cyclic_lr_scheduler_last_epoch,
                },
                "one_cycle_config": {
                    "max_lr": args.one_cycle_lr_scheduler_max_lr,
                    "pct_start": args.one_cycle_lr_scheduler_pct_start,
                    "anneal_strategy": args.one_cycle_lr_scheduler_anneal_strategy,
                    "cycle_momentum": args.one_cycle_lr_scheduler_cycle_momentum,
                    "base_momentum": args.one_cycle_lr_scheduler_base_momentum,
                    "max_momentum": args.one_cycle_lr_scheduler_max_momentum,
                    "div_factor": args.one_cycle_lr_scheduler_div_factor,
                    "final_div_factor": args.one_cycle_lr_scheduler_final_div_factor,
                    "last_epoch": args.one_cycle_lr_scheduler_last_epoch,
                },
                "cosine_annealing_config": {
                    "last_epoch": args.cosine_annealing_lr_scheduler_last_epoch
                },
            },
            "task_scheduler_config": {
                "task_scheduler": args.task_scheduler,
                "sequential_scheduler_config": {
                    "fillup": args.sequential_scheduler_fillup
                },
                "round_robin_scheduler_config": {
                    "fillup": args.round_robin_scheduler_fillup
                },
                "mixed_scheduler_config": {"fillup": args.mixed_scheduler_fillup},
            },
        },
        "logging_config": {
            "counter_unit": args.counter_unit,
            "evaluation_freq": args.evaluation_freq,
            "writer_config": {"writer": args.writer, "verbose": True},
            "checkpointing": args.checkpointing,
            "checkpointer_config": {
                "checkpoint_path": args.checkpoint_path,
                "checkpoint_freq": args.checkpoint_freq,
                "checkpoint_metric": args.checkpoint_metric,
                "checkpoint_task_metrics": args.checkpoint_task_metrics,
                "checkpoint_runway": args.checkpoint_runway,
                "checkpoint_all": args.checkpoint_all,
                "clear_intermediate_checkpoints": args.clear_intermediate_checkpoints,
                "clear_all_checkpoints": args.clear_all_checkpoints,
            },
        },
    }

    return config



from typing import Dict, List, Optional

from numpy import ndarray

from emmental.metrics.precision import precision_scorer
from emmental.metrics.recall import recall_scorer
from emmental.utils.utils import prob_to_pred


def fbeta_scorer(
    golds: ndarray,
    probs: Optional[ndarray],
    preds: ndarray,
    uids: Optional[List[str]] = None,
    pos_label: int = 1,
    beta: int = 1,
) -> Dict[str, float]:
    """F-beta score is the weighted harmonic mean of precision and recall.

    Args:
      golds(ndarray): Ground truth values.
      probs(ndarray or None): Predicted probabilities.
      preds(ndarray): Predicted values.
      uids(list, optional): Unique ids, defaults to None.
      pos_label(int, optional): The positive class label, defaults to 1.
      beta(float, optional): Weight of precision in harmonic mean, defaults to 1.

    Returns:
      dict: F-beta score.

    """

    # Convert probabilistic label to hard label
    if len(golds.shape) == 2:
        golds = prob_to_pred(golds)

    precision = precision_scorer(golds, probs, preds, uids, pos_label)["precision"]
    recall = recall_scorer(golds, probs, preds, uids, pos_label)["recall"]

    fbeta = (
        (1 + beta ** 2) * (precision * recall) / ((beta ** 2 * precision) + recall)
        if (beta ** 2 * precision) + recall > 0
        else 0.0
    )

    return {f"f{beta}": fbeta}


def f1_scorer(
    golds: ndarray,
    probs: Optional[ndarray],
    preds: ndarray,
    uids: Optional[List[str]] = None,
    pos_label: int = 1,
) -> Dict[str, float]:
    """F-1 score.

    Args:
      golds(ndarray): Ground truth values.
      probs(ndarray or None): Predicted probabilities.
      preds(ndarray): Predicted values.
      uids(list, optional): Unique ids.
      pos_label(int, optional): The positive class label, defaults to 1.

    Returns:
      dict: F-1 score.

    """

    return {"f1": fbeta_scorer(golds, probs, preds, uids, pos_label, beta=1)["f1"]}



import logging
from functools import wraps
from types import SimpleNamespace
from typing import Callable, List

import numpy as np
import torch
from numpy import ndarray

from emmental.data import EmmentalDataset

logger = logging.getLogger(__name__)


class slicing_function:
    r"""When wrapped with this decorator, slicing functions only need to return an
      indicator for whether an individual example (bundle of attributes) belongs
      in that slice. Iterating through the dataset, making the pred array (and
      masking), etc. are all handled automatically.

    Args:
      fields(list): Data attributes to use, defaults to [].

    """

    def __init__(self, fields: List[str] = []) -> None:
        self.fields = fields

    def __call__(self, f: Callable) -> Callable:
        @wraps(f)
        def wrapped_f(dataset: EmmentalDataset) -> ndarray:
            """

            Args:
              dataset(EmmentalDataset): Dataset to apply slicing function.

            Returns:
              ndarray: Indicators.

            """
            inds = []
            for idx in range(len(dataset)):
                example = SimpleNamespace(
                    **{field: dataset.X_dict[field][idx] for field in self.fields}
                )
                in_slice = f(example)
                inds.append(1 if in_slice else 0)
            inds = torch.from_numpy(np.array(inds)).view(-1)  # type: ignore
            logger.info(
                f"Total {int((inds == 1).sum())} / {len(dataset)} "  # type: ignore
                f" examples are "
                f"in slice {f.__name__}"
            )
            return inds

        return wrapped_f



import copy
import logging
from functools import partial
from typing import Callable, Dict, List, Optional

from torch import Tensor, nn

from emmental.contrib.slicing.modules import utils
from emmental.contrib.slicing.modules.slice_attention_module import SliceAttentionModule
from emmental.meta import Meta
from emmental.scorer import Scorer
from emmental.task import EmmentalTask
from emmental.utils.utils import move_to_device

logger = logging.getLogger(__name__)


def build_slice_tasks(
    task: EmmentalTask,
    slice_func_dict: Dict[str, Callable],
    slice_scorer: Optional[Scorer] = None,
    slice_distribution: Dict[str, Tensor] = {},
    dropout: float = 0.0,
    slice_ind_head_module: Optional[nn.Module] = None,
    sep_slice_ind_feature: bool = False,
) -> List[EmmentalTask]:
    r"""A function to build slice tasks based on slicing functions.

      We assume the original task flow contains feature extractor and predictor head.
      - The predictor head action should be the last action
      - The feature extractor action should be input of the predictor head action

      For each slicing this function will create two corresponding tasks
      - A slice indicator task to learn whether the data sample is in the slice or not.
      - A slice predictor task that is only learned on the data samples in that slice

      All slice tasks are based on feature extractor module and a slice attention
      module will combine all slice task head to make the final predictions.

    Args:
      task(EmmentalTask): Task to do slicing learning.
      slice_func_dict(dict): Slicing functions.
      slice_scorer(Scorer): Slice scorer, defaults to None.
      slice_distribution(dict): Slice data class distribution, defaults to {}.
      dropout(float): Dropout, defaults to 0.0.
      slice_ind_head_module(nn.Module, optional): Slice indicator head module,
        defaults to None.
      sep_slice_ind_feature(bool): Whether to use sep slice ind feature,
        defaults to False.

    Returns:
      List[EmmentalTask]: list of tasks.

    """

    # Collect task predictor module info
    base_task_predictor_action = task.task_flow[-1]
    base_task_predictor_module = task.module_pool[
        base_task_predictor_action["module"]  # type: ignore
    ]
    if isinstance(base_task_predictor_module, nn.DataParallel):
        base_task_predictor_module = base_task_predictor_module.module

    task_feature_size = base_task_predictor_module.in_features
    task_cardinality = base_task_predictor_module.out_features

    # Remove the predictor head module and action
    base_task_module_pool = task.module_pool
    del base_task_module_pool[base_task_predictor_action["module"]]  # type: ignore

    base_task_task_flow = task.task_flow[:-1]

    tasks = []
    slice_module_pool = nn.ModuleDict()
    for module_name, module in task.module_pool.items():
        slice_module_pool[module_name] = module
    slice_actions = [action for action in base_task_task_flow]

    if slice_ind_head_module is None:
        slice_ind_head_module = nn.Linear(task_feature_size, 2)

    assert isinstance(slice_ind_head_module, nn.Module)

    if slice_scorer is None or not isinstance(slice_scorer, Scorer):
        slice_scorer = Scorer(metrics=["f1"])

    # Create slice indicator tasks.
    # (Note: indicator only has two classes, e.g, in the slice or out)
    for slice_name in slice_func_dict.keys():
        # Create task name
        ind_task_name = f"{task.name}_slice:ind_{slice_name}"

        # Create ind module
        ind_head_module_name = f"{ind_task_name}_head"
        ind_head_module = copy.deepcopy(slice_ind_head_module)

        ind_head_dropout_module_name = f"{task.name}_slice:dropout_{slice_name}"
        ind_head_dropout_module = nn.Dropout(p=dropout)

        # Create module_pool
        ind_module_pool = nn.ModuleDict(
            {
                module_name: module
                for module_name, module in base_task_module_pool.items()
            }
        )
        ind_module_pool[ind_head_dropout_module_name] = ind_head_dropout_module
        ind_module_pool[ind_head_module_name] = ind_head_module

        assert len(base_task_predictor_action["inputs"]) == 1

        ind_head_dropout_module_input_name = base_task_predictor_action["inputs"][0][0]
        ind_head_dropout_module_input_idx = 1 if sep_slice_ind_feature else 0

        # Create task_flow
        ind_task_flow = [action for action in base_task_task_flow]
        ind_task_flow.extend(
            [
                {
                    "name": ind_head_dropout_module_name,
                    "module": ind_head_dropout_module_name,
                    "inputs": [
                        (
                            ind_head_dropout_module_input_name,
                            ind_head_dropout_module_input_idx,
                        )
                    ],
                },
                {
                    "name": ind_head_module_name,
                    "module": ind_head_module_name,
                    "inputs": [(ind_head_dropout_module_name, 0)],
                },
            ]
        )

        # Add slice specific module to slice_module_pool
        slice_module_pool[ind_head_module_name] = ind_head_module
        slice_actions.extend(
            [
                {
                    "name": ind_head_dropout_module_name,
                    "module": ind_head_dropout_module_name,
                    "inputs": [
                        (
                            ind_head_dropout_module_input_name,
                            ind_head_dropout_module_input_idx,
                        )
                    ],
                },
                {
                    "name": ind_head_module_name,
                    "module": ind_head_module_name,
                    "inputs": [(ind_head_dropout_module_name, 0)],
                },
            ]
        )

        # Loss function
        if ind_task_name in slice_distribution:
            loss = partial(
                utils.ce_loss,
                ind_head_module_name,
                weight=move_to_device(
                    slice_distribution[ind_task_name],
                    Meta.config["model_config"]["device"],
                ),
            )
        else:
            loss = partial(utils.ce_loss, ind_head_module_name)

        tasks.append(
            EmmentalTask(
                name=ind_task_name,
                module_pool=ind_module_pool,
                task_flow=ind_task_flow,
                loss_func=loss,
                output_func=partial(utils.output, ind_head_module_name),
                scorer=slice_scorer,
            )
        )

    # Create slice predictor tasks

    # Create share predictor for all slice predictor
    shared_pred_head_module_name = f"{task.name}_slice:shared_pred"
    shared_pred_head_module = nn.Linear(task_feature_size, task_cardinality)

    # Add slice specific module to slice_module_pool
    slice_module_pool[shared_pred_head_module_name] = shared_pred_head_module

    for slice_name in slice_func_dict.keys():
        # Create task name
        pred_task_name = f"{task.name}_slice:pred_{slice_name}"

        # Create pred module
        pred_head_module_name = f"{pred_task_name}_head"
        pred_transform_module_name = f"{task.name}_slice:transform_{slice_name}"
        pred_transform_module = nn.Linear(task_feature_size, task_feature_size)

        # Create module_pool
        pred_module_pool = nn.ModuleDict(
            {
                module_name: module
                for module_name, module in base_task_module_pool.items()
            }
        )
        pred_module_pool[pred_transform_module_name] = pred_transform_module
        pred_module_pool[shared_pred_head_module_name] = shared_pred_head_module

        # Create task_flow
        pred_task_flow = [action for action in base_task_task_flow]
        pred_task_flow.extend(
            [
                {
                    "name": pred_transform_module_name,
                    "module": pred_transform_module_name,
                    "inputs": base_task_predictor_action["inputs"],
                },
                {
                    "name": pred_head_module_name,
                    "module": shared_pred_head_module_name,
                    "inputs": [(pred_transform_module_name, 0)],
                },
            ]
        )

        # Add slice specific module to slice_module_pool
        slice_module_pool[pred_transform_module_name] = pred_transform_module
        slice_actions.extend(
            [
                {
                    "name": pred_transform_module_name,
                    "module": pred_transform_module_name,
                    "inputs": base_task_predictor_action["inputs"],
                },
                {
                    "name": pred_head_module_name,
                    "module": shared_pred_head_module_name,
                    "inputs": [(pred_transform_module_name, 0)],
                },
            ]
        )

        # Loss function
        if pred_task_name in slice_distribution:
            loss = partial(
                utils.ce_loss,
                pred_head_module_name,
                weight=move_to_device(
                    slice_distribution[pred_task_name],
                    Meta.config["model_config"]["device"],
                ),
            )
        else:
            loss = partial(utils.ce_loss, pred_head_module_name)

        tasks.append(
            EmmentalTask(
                name=pred_task_name,
                module_pool=pred_module_pool,
                task_flow=pred_task_flow,
                loss_func=loss,
                output_func=partial(utils.output, pred_head_module_name),
                scorer=task.scorer,
            )
        )

    # Create master task

    # Create task name
    master_task_name = task.name

    # Create attention module
    master_attention_module_name = f"{master_task_name}_attention"
    master_attention_module = SliceAttentionModule(
        slice_ind_key="_slice:ind_",
        slice_pred_key="_slice:pred_",
        slice_pred_feat_key="_slice:transform_",
    )

    # Create module pool
    master_head_module_name = f"{master_task_name}_head"
    master_head_module = base_task_predictor_module

    master_module_pool = slice_module_pool
    master_module_pool[master_attention_module_name] = master_attention_module
    master_module_pool[master_head_module_name] = master_head_module

    # Create task_flow
    master_task_flow = slice_actions + [
        {
            "name": master_attention_module_name,
            "module": master_attention_module_name,
            "inputs": [],  # type: ignore
        },
        {
            "name": master_head_module_name,
            "module": master_head_module_name,
            "inputs": [(master_attention_module_name, 0)],
        },
    ]

    tasks.append(
        EmmentalTask(
            name=master_task_name,
            module_pool=master_module_pool,
            task_flow=master_task_flow,
            loss_func=partial(utils.ce_loss, master_head_module_name),
            output_func=partial(utils.output, master_head_module_name),
            scorer=task.scorer,
        )
    )

    return tasks



import logging
from typing import Dict, Union

from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.optimizer import Optimizer

from emmental import Meta
from emmental.logging.checkpointer import Checkpointer
from emmental.logging.log_writer import LogWriter
from emmental.logging.tensorboard_writer import TensorBoardWriter
from emmental.model import EmmentalModel

logger = logging.getLogger(__name__)


class LoggingManager(object):
    r"""A class to manage logging during training progress.

    Args:
      n_batches_per_epoch(int): Total number batches per epoch.

    """

    def __init__(self, n_batches_per_epoch: int) -> None:
        self.n_batches_per_epoch = n_batches_per_epoch

        # Set up counter

        # Set up evaluation/checkpointing unit (sample, batch, epoch)
        self.counter_unit = Meta.config["logging_config"]["counter_unit"]

        if self.counter_unit not in ["sample", "batch", "epoch"]:
            raise ValueError(f"Unrecognized unit: {self.counter_unit}")

        # Set up evaluation frequency
        self.evaluation_freq = Meta.config["logging_config"]["evaluation_freq"]
        if Meta.config["meta_config"]["verbose"]:
            logger.info(f"Evaluating every {self.evaluation_freq} {self.counter_unit}.")

        if Meta.config["logging_config"]["checkpointing"]:
            self.checkpointing = True

            # Set up checkpointing frequency
            self.checkpointing_freq = int(
                Meta.config["logging_config"]["checkpointer_config"]["checkpoint_freq"]
            )
            if Meta.config["meta_config"]["verbose"]:
                logger.info(
                    f"Checkpointing every "
                    f"{self.checkpointing_freq * self.evaluation_freq} "
                    f"{self.counter_unit}."
                )

            # Set up checkpointer
            self.checkpointer = Checkpointer()
        else:
            self.checkpointing = False
            if Meta.config["meta_config"]["verbose"]:
                logger.info("No checkpointing.")

        # Set up number of samples passed since last evaluation/checkpointing and
        # total number of samples passed since learning process
        self.sample_count: int = 0
        self.sample_total: int = 0

        # Set up number of batches passed since last evaluation/checkpointing and
        # total number of batches passed since learning process
        self.batch_count: int = 0
        self.batch_total: int = 0

        # Set up number of epochs passed since last evaluation/checkpointing and
        # total number of epochs passed since learning process
        self.epoch_count: Union[float, int] = 0
        self.epoch_total: Union[float, int] = 0

        # Set up number of unit passed since last evaluation/checkpointing and
        # total number of unit passed since learning process
        self.unit_count: Union[float, int] = 0
        self.unit_total: Union[float, int] = 0

        # Set up count that triggers the evaluation since last checkpointing
        self.trigger_count = 0

        # Set up log writer
        writer_opt = Meta.config["logging_config"]["writer_config"]["writer"]

        if writer_opt is None:
            self.writer = None
        elif writer_opt == "json":
            self.writer = LogWriter()
        elif writer_opt == "tensorboard":
            self.writer = TensorBoardWriter()
        else:
            raise ValueError(f"Unrecognized writer option '{writer_opt}'")

    def update(self, batch_size: int) -> None:
        r"""Update the counter.

        Args:
          batch_size(int): The number of the samples in the batch.

        """

        # Update number of samples
        self.sample_count += batch_size
        self.sample_total += batch_size

        # Update number of batches
        self.batch_count += 1
        self.batch_total += 1

        # Update number of epochs
        self.epoch_count = self.batch_count / self.n_batches_per_epoch
        self.epoch_total = self.batch_total / self.n_batches_per_epoch

        # Update number of units
        if self.counter_unit == "sample":
            self.unit_count = self.sample_count
            self.unit_total = self.sample_total
        if self.counter_unit == "batch":
            self.unit_count = self.batch_count
            self.unit_total = self.batch_total
        elif self.counter_unit == "epoch":
            self.unit_count = self.epoch_count
            self.unit_total = self.epoch_total

    def trigger_evaluation(self) -> bool:
        r"""Check if triggers the evaluation."""

        satisfied = self.unit_count >= self.evaluation_freq
        if satisfied:
            self.trigger_count += 1
            self.reset()
        return satisfied

    def trigger_checkpointing(self) -> bool:
        r"""Check if triggers the checkpointing."""

        if not self.checkpointing:
            return False
        satisfied = self.trigger_count >= self.checkpointing_freq
        if satisfied:
            self.trigger_count = 0
        return satisfied

    def reset(self) -> None:
        r"""Reset the counter."""

        self.sample_count = 0
        self.batch_count = 0
        self.epoch_count = 0
        self.unit_count = 0

    def write_log(self, metric_dict: Dict[str, float]) -> None:
        r"""Write the metrics to the log.

        Args:
          metric_dict(dict): The metric dict.

        """
        for metric_name, metric_value in metric_dict.items():
            self.writer.add_scalar(metric_name, metric_value, self.unit_total)

    def checkpoint_model(
        self,
        model: EmmentalModel,
        optimizer: Optimizer,
        lr_scheduler: _LRScheduler,
        metric_dict: Dict[str, float],
    ) -> None:
        r"""Checkpoint the model.

        Args:
          model(EmmentalModel): The model to checkpoint.
          optimizer(Optimizer): The optimizer used during training process.
          lr_scheduler(_LRScheduler): Learning rate scheduler.
          metric_dict(dict): the metric dict.

        """

        self.checkpointer.checkpoint(
            self.unit_total, model, optimizer, lr_scheduler, metric_dict
        )

    def close(self, model: EmmentalModel) -> EmmentalModel:
        r"""Close the checkpointer and reload the model if necessary.

        Args:
          model(EmmentalModel): The trained model.

        Returns:
          EmmentalModel: The reloaded model if necessary

        """
        self.writer.close()
        if self.checkpointing:
            model = self.checkpointer.load_best_model(model)
            self.checkpointer.clear()
        return model



from emmental.schedulers.mixed_scheduler import MixedScheduler
from emmental.schedulers.round_robin_scheduler import RoundRobinScheduler
from emmental.schedulers.sequential_scheduler import SequentialScheduler

SCHEDULERS = {
    "mixed": MixedScheduler,
    "round_robin": RoundRobinScheduler,
    "sequential": SequentialScheduler,
}

__all__ = ["mixed", "round_robin", "sequential"]



import itertools
import logging
import os
from collections import defaultdict
from collections.abc import Iterable
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
from numpy import ndarray
from torch import Tensor
from torch.nn import ModuleDict

from emmental.data import EmmentalDataLoader
from emmental.meta import Meta
from emmental.scorer import Scorer
from emmental.task import EmmentalTask
from emmental.utils.utils import construct_identifier, move_to_device, prob_to_pred

logger = logging.getLogger(__name__)


class EmmentalModel(nn.Module):
    r"""A class to build multi-task model.

    Args:
      name(str, optional): Name of the model, defaults to None.
      tasks(EmmentalTask or List[EmmentalTask]): A task or a list of tasks.

    """

    def __init__(
        self,
        name: Optional[str] = None,
        tasks: Optional[Union[EmmentalTask, List[EmmentalTask]]] = None,
    ) -> None:
        super().__init__()
        self.name = name if name is not None else type(self).__name__

        # Initiate the model attributes
        self.module_pool: ModuleDict = ModuleDict()
        self.task_names: Set[str] = set()
        self.task_flows: Dict[str, Any] = dict()  # TODO: make it concrete
        self.loss_funcs: Dict[str, Callable] = dict()
        self.output_funcs: Dict[str, Callable] = dict()
        self.scorers: Dict[str, Scorer] = dict()
        self.weights: Dict[str, float] = dict()

        # Build network with given tasks
        if tasks is not None:
            self.add_tasks(tasks)

        if Meta.config["meta_config"]["verbose"]:
            logger.info(
                f"Created emmental model {self.name} that contains "
                f"task {self.task_names}."
            )

        # Move model to specified device
        self._move_to_device()

    def _move_to_device(self) -> None:
        r"""Move model to specified device."""

        if Meta.config["model_config"]["device"] >= 0:
            if torch.cuda.is_available():
                if Meta.config["meta_config"]["verbose"]:
                    logger.info(
                        f"Moving model to GPU "
                        f"(cuda:{Meta.config['model_config']['device']})."
                    )
                self.to(torch.device(f"cuda:{Meta.config['model_config']['device']}"))
            else:
                if Meta.config["meta_config"]["verbose"]:
                    logger.info("No cuda device available. Switch to cpu instead.")

    def add_tasks(self, tasks: Union[EmmentalTask, List[EmmentalTask]]) -> None:
        r"""Build the MTL network using all tasks.

        Args:
          tasks(EmmentalTask or List[EmmentalTask]): A task or a list of tasks.

        """

        if not isinstance(tasks, Iterable):
            tasks = [tasks]
        for task in tasks:
            self.add_task(task)

    def add_task(self, task: EmmentalTask) -> None:
        r"""Add a single task into MTL network.

        Args:
          task(EmmentalTask): A task to add.

        """
        if not isinstance(task, EmmentalTask):
            raise ValueError(f"Unrecognized task type {task}.")

        if task.name in self.task_names:
            raise ValueError(
                f"Found duplicate task {task.name}, different task should use "
                f"different task name."
            )

        # Combine module_pool from all tasks
        for key in task.module_pool.keys():
            if key in self.module_pool.keys():
                if Meta.config["model_config"]["dataparallel"]:
                    task.module_pool[key] = nn.DataParallel(self.module_pool[key])
                else:
                    task.module_pool[key] = self.module_pool[key]
            else:
                if Meta.config["model_config"]["dataparallel"]:
                    self.module_pool[key] = nn.DataParallel(task.module_pool[key])
                else:
                    self.module_pool[key] = task.module_pool[key]
        # Collect task name
        self.task_names.add(task.name)
        # Collect task flow
        self.task_flows[task.name] = task.task_flow
        # Collect loss function
        self.loss_funcs[task.name] = task.loss_func
        # Collect output function
        self.output_funcs[task.name] = task.output_func
        # Collect scorer
        self.scorers[task.name] = task.scorer
        # Collect weight
        self.weights[task.name] = task.weight

        # Move model to specified device
        self._move_to_device()

    def update_task(self, task: EmmentalTask) -> None:
        r"""Update a existing task in MTL network

        Args:
          task(EmmentalTask): A task to update.

        """

        # Update module_pool with task
        for key in task.module_pool.keys():
            # Update the model's module with the task's module
            if Meta.config["model_config"]["dataparallel"]:
                self.module_pool[key] = nn.DataParallel(task.module_pool[key])
            else:
                self.module_pool[key] = task.module_pool[key]
        # Update task flow
        self.task_flows[task.name] = task.task_flow
        # Update loss function
        self.loss_funcs[task.name] = task.loss_func
        # Update output function
        self.output_funcs[task.name] = task.output_func
        # Collect scorer
        self.scorers[task.name] = task.scorer
        # Collect weight
        self.weights[task.name] = task.weight

        # Move model to specified device
        self._move_to_device()

    def remove_task(self, task_name: str) -> None:
        r"""Remove a existing task from MTL network

        Args:
          task_name(str): The task name to remove.

        """
        if task_name not in self.task_flows:
            if Meta.config["meta_config"]["verbose"]:
                logger.info(f"Task ({task_name}) not in the current model, skip...")
            return

        # Remove task by task_name
        if Meta.config["meta_config"]["verbose"]:
            logger.info(f"Removing Task {task_name}.")

        self.task_names.remove(task_name)
        del self.task_flows[task_name]
        del self.loss_funcs[task_name]
        del self.output_funcs[task_name]
        del self.scorers[task_name]
        del self.weights[task_name]
        # TODO: remove the modules only associate with that task

    def __repr__(self) -> str:
        cls_name = type(self).__name__
        return f"{cls_name}(name={self.name})"

    def flow(self, X_dict: Dict[str, Any], task_names: List[str]) -> Dict[str, Any]:
        r"""Forward based on input and task flow.

        Note:
          We assume that all shared modules from all tasks are based on the
          same input.

        Args:
          X_dict(dict): The input data
          task_names(list): The task names that needs to forward.

        Returns:
          dict: The output of all forwarded modules

        """

        X_dict = move_to_device(X_dict, Meta.config["model_config"]["device"])

        output_dict = dict(_input_=X_dict)

        # Call forward for each task
        for task_name in task_names:
            for action in self.task_flows[task_name]:
                if action["name"] not in output_dict:
                    if action["inputs"]:
                        try:
                            input = [
                                output_dict[action_name][output_index]
                                for action_name, output_index in action["inputs"]
                            ]
                        except Exception:
                            raise ValueError(f"Unrecognized action {action}.")
                        output = self.module_pool[action["module"]].forward(*input)
                    else:
                        output = self.module_pool[action["module"]].forward(output_dict)
                    if isinstance(output, tuple):
                        output = list(output)
                    if not isinstance(output, list):
                        output = [output]
                    output_dict[action["name"]] = output

        return output_dict

    def forward(  # type: ignore
        self,
        uids: List[str],
        X_dict: Dict[str, Any],
        Y_dict: Dict[str, Tensor],
        task_to_label_dict: Dict[str, str],
    ) -> Tuple[
        Dict[str, List[str]], Dict[str, ndarray], Dict[str, ndarray], Dict[str, ndarray]
    ]:
        r"""Forward function.

        Args:
          uids(list): The uids of input data.
          X_dict(dict): The input data.
          Y_dict(dict): The output data.
          task_to_label_dict(dict): The task to label mapping.

        Returns:
          tuple: The (active) uids, loss and prob in the batch of all tasks.

        """

        uid_dict: Dict[str, List[str]] = defaultdict(list)
        loss_dict: Dict[str, ndarray] = defaultdict(float)
        gold_dict: Dict[str, ndarray] = defaultdict(list)
        prob_dict: Dict[str, ndarray] = defaultdict(list)

        output_dict = self.flow(X_dict, list(task_to_label_dict.keys()))

        # Calculate loss for each task
        for task_name, label_name in task_to_label_dict.items():
            Y = Y_dict[label_name]

            # Select the active samples
            if Meta.config["learner_config"]["ignore_index"] is not None:
                if len(Y.size()) == 1:
                    active = Y.detach() != Meta.config["learner_config"]["ignore_index"]
                else:
                    active = torch.any(
                        Y.detach() != Meta.config["learner_config"]["ignore_index"],
                        dim=1,
                    )
            else:
                active = torch.BoolTensor([True] * Y.size()[0])  # type: ignore

            # Only calculate the loss when active example exists
            if active.any():
                uid_dict[task_name] = [*itertools.compress(uids, active.numpy())]

                loss_dict[task_name] = self.loss_funcs[task_name](
                    output_dict,
                    move_to_device(
                        Y_dict[label_name], Meta.config["model_config"]["device"]
                    ),
                    move_to_device(active, Meta.config["model_config"]["device"]),
                )

                prob_dict[task_name] = (
                    self.output_funcs[task_name](output_dict)[
                        move_to_device(active, Meta.config["model_config"]["device"])
                    ]
                    .cpu()
                    .detach()
                    .numpy()
                )

                gold_dict[task_name] = Y_dict[label_name][active].cpu().numpy()

        return uid_dict, loss_dict, prob_dict, gold_dict

    @torch.no_grad()
    def predict(
        self, dataloader: EmmentalDataLoader, return_preds: bool = False
    ) -> Dict[str, Any]:
        r"""Predict from dataloader.

        Args:
          dataloader(EmmentalDataLoader): The dataloader to predict.
          return_preds(bool): Whether return predictions or not, defaults to False.

        Returns:
          dict: The result dict.

        """

        self.eval()

        uid_dict: Dict[str, List[str]] = defaultdict(list)
        gold_dict: Dict[str, List[Union[ndarray, int, float]]] = defaultdict(list)
        prob_dict: Dict[str, List[Union[ndarray, int, float]]] = defaultdict(list)
        pred_dict: Dict[str, List[ndarray]] = defaultdict(list)
        # Fix it later
        loss_dict: Dict[str, Union[ndarray, float]] = defaultdict(list)  # type: ignore

        # Collect dataloader information
        task_to_label_dict = dataloader.task_to_label_dict
        uid = dataloader.uid

        for batch_num, (X_bdict, Y_bdict) in enumerate(dataloader):
            uid_bdict, loss_bdict, prob_bdict, gold_bdict = self.forward(
                X_bdict[uid], X_bdict, Y_bdict, task_to_label_dict
            )
            for task_name in uid_bdict.keys():
                uid_dict[task_name].extend(uid_bdict[task_name])
                prob_dict[task_name].extend(prob_bdict[task_name])
                gold_dict[task_name].extend(gold_bdict[task_name])
                if len(loss_bdict[task_name].size()) == 0:
                    if loss_dict[task_name] == []:
                        loss_dict[task_name] = 0
                    loss_dict[task_name] += loss_bdict[task_name].item() * len(
                        uid_bdict[task_name]
                    )
                else:
                    loss_dict[task_name].extend(  # type: ignore
                        loss_bdict[task_name].cpu().numpy()
                    )

        # Calculate average loss
        for task_name in uid_dict.keys():
            if not isinstance(loss_dict[task_name], list):
                loss_dict[task_name] /= len(uid_dict[task_name])

        res = {
            "uids": uid_dict,
            "golds": gold_dict,
            "probs": prob_dict,
            "losses": loss_dict,
        }

        if return_preds:
            for task_name, prob in prob_dict.items():
                pred_dict[task_name] = prob_to_pred(prob)
            res["preds"] = pred_dict

        return res

    @torch.no_grad()
    def score(
        self,
        dataloaders: Union[EmmentalDataLoader, List[EmmentalDataLoader]],
        return_average: bool = True,
    ) -> Dict[str, float]:
        """Score the data from dataloader.

        Args:
          dataloaders(EmmentalDataLoader or List[EmmentalDataLoader]): The dataloaders
            to score.
          return_average(bool): Whether to return average score.

        Returns:
          dict: Score dict.

        """

        self.eval()

        if not isinstance(dataloaders, list):
            dataloaders = [dataloaders]

        metric_score_dict = dict()

        if return_average:
            micro_score_dict: defaultdict = defaultdict(list)
            macro_score_dict: defaultdict = defaultdict(list)
            macro_loss_dict: defaultdict = defaultdict(list)

        for dataloader in dataloaders:
            predictions = self.predict(dataloader, return_preds=True)
            for task_name in predictions["golds"].keys():
                metric_score = self.scorers[task_name].score(
                    predictions["golds"][task_name],
                    predictions["probs"][task_name],
                    predictions["preds"][task_name],
                    predictions["uids"][task_name],
                )
                for metric_name, metric_value in metric_score.items():
                    identifier = construct_identifier(
                        task_name, dataloader.data_name, dataloader.split, metric_name
                    )
                    metric_score_dict[identifier] = metric_value

                # Store the loss
                identifier = construct_identifier(
                    task_name, dataloader.data_name, dataloader.split, "loss"
                )
                metric_score_dict[identifier] = np.mean(
                    predictions["losses"][task_name]
                )

                if return_average:
                    # Collect average score
                    identifier = construct_identifier(
                        task_name, dataloader.data_name, dataloader.split, "average"
                    )
                    metric_score_dict[identifier] = np.mean(list(metric_score.values()))

                    micro_score_dict[dataloader.split].extend(
                        list(metric_score.values())
                    )
                    macro_score_dict[dataloader.split].append(
                        metric_score_dict[identifier]
                    )

                    # Store the loss
                    identifier = construct_identifier(
                        task_name, dataloader.data_name, dataloader.split, "loss"
                    )
                    macro_loss_dict[dataloader.split].append(
                        metric_score_dict[identifier]
                    )

        if return_average:
            # Collect split-wise micro/macro average score
            for split in micro_score_dict.keys():
                identifier = construct_identifier(
                    "model", "all", split, "micro_average"
                )
                metric_score_dict[identifier] = np.mean(micro_score_dict[split])
                identifier = construct_identifier(
                    "model", "all", split, "macro_average"
                )
                metric_score_dict[identifier] = np.mean(macro_score_dict[split])
                identifier = construct_identifier("model", "all", split, "loss")
                metric_score_dict[identifier] = np.mean(macro_loss_dict[split])

            # Collect overall micro/macro average score/loss
            identifier = construct_identifier("model", "all", "all", "micro_average")
            metric_score_dict[identifier] = np.mean(
                list(itertools.chain.from_iterable(micro_score_dict.values()))
            )
            identifier = construct_identifier("model", "all", "all", "macro_average")
            metric_score_dict[identifier] = np.mean(
                list(itertools.chain.from_iterable(macro_score_dict.values()))
            )
            identifier = construct_identifier("model", "all", "all", "loss")
            metric_score_dict[identifier] = np.mean(
                list(itertools.chain.from_iterable(macro_loss_dict.values()))
            )

        # TODO: have a better to handle global evaluation metric
        if Meta.config["learner_config"]["global_evaluation_metric_dict"]:
            global_evaluation_metric_dict = Meta.config["learner_config"][
                "global_evaluation_metric_dict"
            ]
            for metric_name, metric in global_evaluation_metric_dict.items():
                metric_score_dict[metric_name] = metric(metric_score_dict)

        return metric_score_dict

    def save(self, model_path: str) -> None:
        r"""Save the current model.

        Args:
          model_path(str): Saved model path.

        """

        # Check existence of model saving directory and create if does not exist.
        if not os.path.exists(os.path.dirname(model_path)):
            os.makedirs(os.path.dirname(model_path))

        state_dict = {
            "model": {
                "name": self.name,
                "module_pool": self.collect_state_dict(),
                # "task_names": self.task_names,
                # "task_flows": self.task_flows,
                # "loss_funcs": self.loss_funcs,
                # "output_funcs": self.output_funcs,
                # "scorers": self.scorers,
            }
        }

        try:
            torch.save(state_dict, model_path)
        except BaseException:
            logger.warning("Saving failed... continuing anyway.")

        if Meta.config["meta_config"]["verbose"]:
            logger.info(f"[{self.name}] Model saved in {model_path}")

    def load(self, model_path: str) -> None:
        r"""Load model state_dict from file and reinitialize the model weights.

        Args:
          model_path(str): Saved model path.

        """

        if not os.path.exists(model_path):
            logger.error("Loading failed... Model does not exist.")

        try:
            checkpoint = torch.load(model_path, map_location=torch.device("cpu"))
        except BaseException:
            logger.error(f"Loading failed... Cannot load model from {model_path}")
            raise

        self.load_state_dict(checkpoint["model"]["module_pool"])

        if Meta.config["meta_config"]["verbose"]:
            logger.info(f"[{self.name}] Model loaded from {model_path}")

        # Move model to specified device
        self._move_to_device()

    def collect_state_dict(self) -> Dict[str, Any]:
        r"""Collect the state dict."""

        state_dict: Dict[str, Any] = defaultdict(list)

        for module_name, module in self.module_pool.items():
            if Meta.config["model_config"]["dataparallel"]:
                state_dict[module_name] = module.module.state_dict()  # type: ignore
            else:
                state_dict[module_name] = module.state_dict()

        return state_dict

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:  # type: ignore
        r"""Load the state dict.

        Args:
          state_dict(dict): The state dict to load.

        """

        for module_name, module_state_dict in state_dict.items():
            if module_name in self.module_pool:
                if Meta.config["model_config"]["dataparallel"]:
                    self.module_pool[module_name].module.load_state_dict(
                        module_state_dict
                    )
                else:
                    self.module_pool[module_name].load_state_dict(module_state_dict)
            else:
                logger.info(f"Missing {module_name} in module_pool, skip it..")



import logging
from typing import Callable, Dict, List, Tuple, Union

from torch import nn
from torch.nn.modules.container import ModuleDict

from emmental.meta import Meta
from emmental.scorer import Scorer

logger = logging.getLogger(__name__)


class EmmentalTask(object):
    r"""Task class to define task in Emmental model.

    Args:
      name(str): The name of the task (Primary key).
      module_pool(ModuleDict): A dict of modules that uses in the task.
      task_flow(list): The task flow among modules to define how the data flows.
      loss_func(callable): The function to calculate the loss.
      output_func(callable): The function to generate the output.
      scorer(Scorer): The class of metrics to evaluate the task.
      weight(float or int): The weight of the task.

    """

    def __init__(
        self,
        name: str,
        module_pool: ModuleDict,
        task_flow: List[
            Dict[str, Union[str, List[Tuple[str, str]], List[Tuple[str, int]]]]
        ],
        loss_func: Callable,
        output_func: Callable,
        scorer: Scorer,
        weight: Union[float, int] = 1.0,
    ) -> None:

        self.name = name
        assert isinstance(module_pool, nn.ModuleDict) is True
        self.module_pool = module_pool
        self.task_flow = task_flow
        self.loss_func = loss_func
        self.output_func = output_func
        self.scorer = scorer
        self.weight = weight

        if Meta.config["meta_config"]["verbose"]:
            logger.info(f"Created task: {self.name}")

    def __repr__(self) -> str:
        cls_name = type(self).__name__
        return f"{cls_name}(name={self.name})"



from typing import Dict, List, Optional

import numpy as np
from numpy import ndarray

from emmental.metrics.pearson_correlation import pearson_correlation_scorer
from emmental.metrics.spearman_correlation import spearman_correlation_scorer


def pearson_spearman_scorer(
    golds: ndarray,
    probs: ndarray,
    preds: Optional[ndarray],
    uids: Optional[List[str]] = None,
) -> Dict[str, float]:
    r"""Average of Pearson correlation coefficient and Spearman rank-order
    correlation coefficient.

    Args:
      golds(ndarray): Ground truth values.
      probs(ndarray): Predicted probabilities.
      preds(ndarray or None): Predicted values.
      uids(list, optional): Unique ids, defaults to None.

    Returns:
      dict: The average of Pearson correlation coefficient and Spearman rank-order
      correlation coefficient.

    """

    metrics = dict()
    pearson_correlation = pearson_correlation_scorer(golds, probs, preds, uids)
    spearman_correlation = spearman_correlation_scorer(golds, probs, preds, uids)
    metrics["pearson_spearman"] = np.mean(
        [
            pearson_correlation["pearson_correlation"],
            spearman_correlation["spearman_correlation"],
        ]
    )
    return metrics
